"""
Verify that default behaviors of deepxromm remains the same
"""

import os
from pathlib import Path
import shutil
import unittest

import cv2
import pandas as pd
from datetime import datetime as dt
import yaml

from deepxromm import DeepXROMM

SAMPLE_FRAME = Path(__file__).parent / "sample_frame.jpg"
SAMPLE_FRAME_INPUT = Path(__file__).parent / "sample_frame_input.csv"
SAMPLE_AUTOCORRECT_OUTPUT = Path(__file__).parent / "sample_autocorrect_output.csv"

DEEPXROMM_TEST_CODEC = os.environ.get("DEEPXROMM_TEST_CODEC", "avc1")


class TestDefaultsPerformance(unittest.TestCase):
    """Test that the config will still be configured properly if the user only provides XMAlab input"""

    def setUp(self):
        """Create a sample project where the user only inputs XMAlab data"""
        self.working_dir = Path.cwd() / "tmp"
        self.config = self.working_dir / "project_config.yaml"
        self.deepxromm = DeepXROMM.create_new_project(
            self.working_dir, codec=DEEPXROMM_TEST_CODEC
        )
        frame = cv2.imread(str(SAMPLE_FRAME))

        # Make a trial directory
        (self.working_dir / "trainingdata/dummy").mkdir(parents=True, exist_ok=True)

        # Cam 1
        video_path_1 = self.working_dir / "trainingdata/dummy/dummy_cam1.avi"
        out = cv2.VideoWriter(
            str(video_path_1), cv2.VideoWriter_fourcc(*"DIVX"), 15, (1024, 512)
        )
        out.write(frame)
        out.release()

        # Cam 2
        video_path_2 = self.working_dir / "trainingdata/dummy/dummy_cam2.avi"
        out = cv2.VideoWriter(
            str(video_path_2), cv2.VideoWriter_fourcc(*"DIVX"), 15, (1024, 512)
        )
        out.write(frame)
        out.release()

        # CSV
        df = pd.DataFrame(
            {
                "foo_cam1_X": 0,
                "foo_cam1_Y": 0,
                "foo_cam2_X": 0,
                "foo_cam2_Y": 0,
                "bar_cam1_X": 0,
                "bar_cam1_Y": 0,
                "bar_cam2_X": 0,
                "bar_cam2_Y": 0,
                "baz_cam1_X": 0,
                "baz_cam1_Y": 0,
                "baz_cam2_X": 0,
                "baz_cam2_Y": 0,
            },
            index=[1],
        )
        csv_path = self.working_dir / "trainingdata/dummy/dummy.csv"
        df.to_csv(str(csv_path), index=False)
        cv2.destroyAllWindows()

    def test_can_find_frames_from_csv(self):
        """Can I accurately find the number of frames in the video if the user doesn't tell me?"""
        print(list(self.working_dir.iterdir()))
        deepxromm = DeepXROMM.load_project(self.working_dir)
        config = deepxromm.config
        self.assertEqual(
            config["nframes"], 1, msg=f"Actual nframes: {config['nframes']}"
        )

    def test_analyze_errors_if_no_folders_in_trials_dir(self):
        """If there are no trials to analyze, do we return an error?"""
        with self.assertRaises(FileNotFoundError):
            self.deepxromm.analyze_videos()

    def test_autocorrect_errors_if_no_folders_in_trials_dir(self):
        """If there are no trials to autocorrect, do we return an error?"""
        with self.assertRaises(FileNotFoundError):
            self.deepxromm.autocorrect_trials()

    def test_warn_users_if_nframes_doesnt_match_csv(self):
        """If the number of frames in the CSV doesn't match the number of frames specified, do I issue a warning?"""
        with self.config.open() as config:
            tmp = yaml.safe_load(config)

        # Modify the number of frames (similar to how a user would)
        tmp["nframes"] = 2
        with self.config.open("w") as fp:
            yaml.dump(tmp, fp, sort_keys=False)

        # Check that the user is warned
        with self.assertWarns(UserWarning):
            DeepXROMM.load_project(self.working_dir)

    def test_yaml_file_updates_nframes_after_load_if_frames_is_0(self):
        """If the user doesn't specify how many frames they want analyzed,
        does their YAML file get updated with how many are in the CSV?"""
        DeepXROMM.load_project(self.working_dir)
        with self.config.open() as config:
            tmp = yaml.safe_load(config)
        self.assertEqual(tmp["nframes"], 1, msg=f"Actual nframes: {tmp['nframes']}")

    def test_warn_if_user_has_tracked_less_than_threshold_frames(self):
        """If the user has tracked less than threshold % of their trial,
        do I give them a warning?"""
        DeepXROMM.load_project(self.working_dir)

        # Increase the number of frames in the video to 100 so I can test this
        frame = cv2.imread(str(SAMPLE_FRAME))
        video_path = self.working_dir / "trainingdata/dummy/dummy_cam1.avi"
        out = cv2.VideoWriter(
            str(video_path), cv2.VideoWriter_fourcc(*"DIVX"), 15, (1024, 512)
        )
        for _ in range(100):
            out.write(frame)
        out.release()

        # Check that the user is warned
        with self.assertWarns(UserWarning):
            DeepXROMM.load_project(self.working_dir)

    def test_bodyparts_add_from_csv_if_not_defined(self):
        """If the user hasn't specified the bodyparts from their trial, we can pull them from the CSV"""
        date = dt.today().strftime("%Y-%m-%d")
        DeepXROMM.load_project(self.working_dir)
        config_path = Path(self.working_dir) / f"tmp-NA-{date}" / "config.yaml"

        with config_path.open() as dlc_config:
            config_obj = yaml.safe_load(dlc_config)
        self.assertEqual(config_obj["bodyparts"], ["foo", "bar", "baz"])

    def test_bodyparts_add_from_csv_in_3d(self):
        """If the user wants to do 3D tracking, we output the desired list of bodyparts"""
        date = dt.today().strftime("%Y-%m-%d")
        with self.config.open("r") as config:
            tmp = yaml.safe_load(config)
        tmp["mode"] = "rgb"
        with self.config.open("w") as fp:
            yaml.dump(tmp, fp, sort_keys=False)
        DeepXROMM.load_project(self.working_dir)

        path_to_config = self.working_dir / f"tmp-NA-{date}" / "config.yaml"
        with path_to_config.open("r") as dlc_config:
            config_obj = yaml.safe_load(dlc_config)

        self.assertEqual(
            config_obj["bodyparts"],
            ["foo_cam1", "foo_cam2", "bar_cam1", "bar_cam2", "baz_cam1", "baz_cam2"],
        )

    def test_bodyparts_add_synthetic(self):
        """Can we add swapped markers?"""
        date = dt.today().strftime("%Y-%m-%d")

        with self.config.open("r") as config:
            tmp = yaml.safe_load(config)
        tmp["mode"] = "rgb"
        tmp["swapped_markers"] = True
        with self.config.open("w") as fp:
            yaml.dump(tmp, fp, sort_keys=False)
        DeepXROMM.load_project(self.working_dir)

        path_to_config = self.working_dir / f"tmp-NA-{date}" / "config.yaml"
        with path_to_config.open("r") as dlc_config:
            config_obj = yaml.safe_load(dlc_config)

        self.assertEqual(
            config_obj["bodyparts"],
            [
                "foo_cam1",
                "foo_cam2",
                "bar_cam1",
                "bar_cam2",
                "baz_cam1",
                "baz_cam2",
                "sw_foo_cam1",
                "sw_foo_cam2",
                "sw_bar_cam1",
                "sw_bar_cam2",
                "sw_baz_cam1",
                "sw_baz_cam2",
            ],
        )

    def test_bodyparts_add_crossed(self):
        """Can we add crossed markers?"""
        date = dt.today().strftime("%Y-%m-%d")

        with self.config.open("r") as config:
            tmp = yaml.safe_load(config)
        tmp["mode"] = "rgb"
        tmp["crossed_markers"] = True
        with self.config.open("w") as fp:
            yaml.dump(tmp, fp, sort_keys=False)
        DeepXROMM.load_project(self.working_dir)

        path_to_config = self.working_dir / f"tmp-NA-{date}" / "config.yaml"
        with path_to_config.open("r") as dlc_config:
            config_obj = yaml.safe_load(dlc_config)

        self.assertEqual(
            config_obj["bodyparts"],
            [
                "foo_cam1",
                "foo_cam2",
                "bar_cam1",
                "bar_cam2",
                "baz_cam1",
                "baz_cam2",
                "cx_foo_cam1x2",
                "cx_bar_cam1x2",
                "cx_baz_cam1x2",
            ],
        )

    def test_bodyparts_add_synthetic_and_crossed(self):
        """Can we add both swapped and crossed markers?"""
        date = dt.today().strftime("%Y-%m-%d")

        with self.config.open("r") as config:
            tmp = yaml.safe_load(config)
        tmp["mode"] = "rgb"
        tmp["swapped_markers"] = True
        tmp["crossed_markers"] = True
        with self.config.open("w") as fp:
            yaml.dump(tmp, fp, sort_keys=False)
        DeepXROMM.load_project(self.working_dir)

        path_to_config = self.working_dir / f"tmp-NA-{date}" / "config.yaml"
        with path_to_config.open("r") as dlc_config:
            config_obj = yaml.safe_load(dlc_config)

        self.assertEqual(
            config_obj["bodyparts"],
            [
                "foo_cam1",
                "foo_cam2",
                "bar_cam1",
                "bar_cam2",
                "baz_cam1",
                "baz_cam2",
                "sw_foo_cam1",
                "sw_foo_cam2",
                "sw_bar_cam1",
                "sw_bar_cam2",
                "sw_baz_cam1",
                "sw_baz_cam2",
                "cx_foo_cam1x2",
                "cx_bar_cam1x2",
                "cx_baz_cam1x2",
            ],
        )

    def test_bodyparts_error_if_different_from_csv(self):
        """If the user specifies different bodyparts than their CSV, raise an error"""
        date = dt.today().strftime("%Y-%m-%d")
        path_to_config = self.working_dir / f"tmp-NA-{date}" / "config.yaml"
        with path_to_config.open("r") as dlc_config:
            config_dlc = yaml.safe_load(dlc_config)

        config_dlc["bodyparts"] = ["foo", "bar"]

        with path_to_config.open("w") as dlc_config:
            yaml.dump(config_dlc, dlc_config, sort_keys=False)

        with self.assertRaises(SyntaxError):
            DeepXROMM.load_project(self.working_dir)

    def test_autocorrect_error_if_trial_not_set(self):
        """If the user doesn't specify a trial to test autocorrect with, do we error?"""
        with self.config.open("r") as config:
            tmp = yaml.safe_load(config)

        tmp["test_autocorrect"] = True
        tmp["marker"] = "foo"
        with self.config.open("w") as fp:
            yaml.dump(tmp, fp, sort_keys=False)

        with self.assertRaises(SyntaxError):
            DeepXROMM.load_project(self.working_dir)

    def test_autocorrect_error_if_marker_not_set(self):
        """If the user doesn't specify a marker to test autocorrect with, do we error?"""
        with self.config.open("r") as config:
            tmp = yaml.safe_load(config)

        tmp["test_autocorrect"] = True
        tmp["trial_name"] = "test"

        with self.config.open("w") as fp:
            yaml.dump(tmp, fp, sort_keys=False)

        with self.assertRaises(SyntaxError):
            DeepXROMM.load_project(self.working_dir)

    def test_config_update_does_not_sort_keys(self):
        """When we update the config, does the data remain sorted according to how it is in default_config?"""
        with open(self.working_dir / "project_config.yaml", "r") as f:
            current_config = yaml.safe_load(f)
        DeepXROMM.load_project(self.working_dir)
        with open(self.working_dir / "project_config.yaml", "r") as f:
            updated_config = yaml.safe_load(f)

        for current_key, updated_key in zip(
            current_config.keys(), updated_config.keys()
        ):
            with self.subTest(folder=current_key):
                self.assertTrue(current_key == updated_key)

    def test_migration_from_tracking_mode_to_mode(self):
        """Does loading a config with deprecated 'tracking_mode' auto-migrate to 'mode'?"""
        # Modify config to use deprecated key
        with self.config.open("r") as config:
            tmp = yaml.safe_load(config)
        del tmp["mode"]  # Remove new key if it exists
        tmp["tracking_mode"] = "2D"  # Use deprecated key
        with self.config.open("w") as fp:
            yaml.dump(tmp, fp, sort_keys=False)

        # Load project (should trigger migration)
        with self.assertWarns(DeprecationWarning):
            DeepXROMM.load_project(self.working_dir)

        # Verify config was migrated and saved
        with self.config.open("r") as config:
            migrated = yaml.safe_load(config)

        self.assertIn("mode", migrated)
        self.assertNotIn("tracking_mode", migrated)
        self.assertEqual(migrated["mode"], "2D")

    def test_conflicting_mode_values_raises_error(self):
        """Does having both 'mode' and 'tracking_mode' with different values raise ValueError?"""
        # Create config with conflicting values
        with self.config.open("r") as config:
            tmp = yaml.safe_load(config)
        tmp["mode"] = "2D"
        tmp["tracking_mode"] = "rgb"  # Different value
        with self.config.open("w") as fp:
            yaml.dump(tmp, fp, sort_keys=False)

        # Verify error is raised
        with self.assertRaises(ValueError) as context:
            DeepXROMM.load_project(self.working_dir)

        error_message = str(context.exception)
        self.assertIn("Conflicting values", error_message)
        self.assertIn("mode", error_message)
        self.assertIn("tracking_mode", error_message)

    def test_duplicate_mode_values_migrates_successfully(self):
        """Does having both keys with same value migrate successfully?"""
        # Create config with duplicate but matching values
        with self.config.open("r") as config:
            tmp = yaml.safe_load(config)
        tmp["mode"] = "2D"
        tmp["tracking_mode"] = "2D"  # Same value
        with self.config.open("w") as fp:
            yaml.dump(tmp, fp, sort_keys=False)

        # Load project (should trigger migration with warning)
        with self.assertWarns(DeprecationWarning):
            DeepXROMM.load_project(self.working_dir)

        # Verify only 'mode' remains
        with self.config.open("r") as config:
            migrated = yaml.safe_load(config)

        self.assertIn("mode", migrated)
        self.assertNotIn("tracking_mode", migrated)
        self.assertEqual(migrated["mode"], "2D")

    def tearDown(self):
        """Remove the created temp project"""
        shutil.rmtree(self.working_dir)



"""Primary interface for analyzing the XROMM data using trained network"""

import os
import math
import logging
from pathlib import Path

from itertools import combinations
import cv2
import deeplabcut
import imagehash
import pandas as pd
from PIL import Image
import yaml

from .xma_data_processor import XMADataProcessor

logger = logging.getLogger(__name__)
logging.basicConfig(
    filename="analyzer.log",
    encoding="utf-8",
    level=os.environ.get("DEEPXROMM_LOG_LEVEL", "INFO").upper(),
)


class Analyzer:
    """Analyzes XROMM videos using trained network."""

    def __init__(self, config):
        self.working_dir = Path(config["working_dir"])
        self.project_config = self.working_dir / "project_config.yaml"
        self._trials_path = self.working_dir / "trials"
        self._data_processor = XMADataProcessor(config)
        self._config = config
        self._dlc_config = config["path_config_file"]

    def analyze_videos(self):
        """Analyze videos with a pre-existing network"""
        trials = self._data_processor.list_trials()

        # Establish project vars
        with open(self._dlc_config) as dlc_config:
            dlc = yaml.safe_load(dlc_config)
        iteration = dlc["iteration"]

        mode = self._config["mode"]
        if mode in ["2D", "per_cam"]:
            self._analyze_xromm_videos(iteration)

        elif mode == "rgb":
            self._data_processor.make_rgb_videos("trials")
            for trial_path in trials:
                trial = trial_path.name
                current_files = trial_path.glob("*")
                logger.debug(f"Current files in directory {current_files}")
                video_path = trial_path / f"{trial}_rgb.avi"
                destfolder = trial_path / f"it{iteration}"
                deeplabcut.analyze_videos(
                    self._dlc_config,
                    str(
                        video_path
                    ),  # DLC relies on .endswith to determine suffix, so this needs to be a string
                    destfolder=destfolder,
                    save_as_csv=True,
                )
        else:
            raise ValueError(f"Unsupported mode: {mode}")

    def _analyze_xromm_videos(self, iteration: int) -> None:
        """Analyze all novel videos in the 'trials' folder of a deepxromm project"""
        # assumes you have cam1 and cam2 videos as .avi in their own seperate trial folders
        # assumes all folders w/i new_data_path are trial folders

        # analyze videos
        cameras = [1, 2]
        trials = self._data_processor.list_trials()
        mode = self._config["mode"]

        for trialpath in trials:
            savepath = trialpath / f"it{iteration}"
            if savepath.exists():
                temp = savepath.glob("*Predicted2DPoints.csv")
                if temp:
                    print(
                        f"There are already predicted points in iteration {iteration} subfolders... skipping point prediction"
                    )
                    return
            else:
                savepath.mkdir(parents=True, exist_ok=True)  # make new folder
            # get video file
            for camera in cameras:
                # Error handling handled by find_cam_file helper
                video = self._data_processor.find_cam_file(trialpath, f"cam{camera}")
                # analyze video
                if mode == "2D":
                    deeplabcut.analyze_videos(
                        self._config["path_config_file"],
                        [
                            str(video)
                        ],  # DLC uses endswith filtering for suffixes for some reason
                        destfolder=savepath,
                        save_as_csv=True,
                    )
                else:
                    configs = [
                        self._config["path_config_file"],
                        self._config["path_config_file_2"],
                    ]
                    deeplabcut.analyze_videos(
                        configs[camera - 1],
                        [
                            str(video)
                        ],  # DLC uses endswith filtering for suffixes for some reason
                        destfolder=savepath,
                        save_as_csv=True,
                    )

    def analyze_video_similarity_project(self):
        """Analyze all videos in a project and take their average similar. This is dangerous, as it will assume that all cam1/cam2 pairs match
        or don't match!"""
        similarity_score = {}
        trial_combos = combinations(self._data_processor.list_trials(), 2)
        for trial1, trial2 in trial_combos:
            self._config["trial_1_name"] = trial1.stem
            self._config["trial_2_name"] = trial2.stem
            with self.project_config.open("w") as file:
                yaml.dump(self._config, file, sort_keys=False)
            # TODO: pass the trial names directly instead of writing into config
            similarity_score[(trial1, trial2)] = self.analyze_video_similarity_trial()
        return similarity_score

    def analyze_video_similarity_trial(self):
        """Analyze the average similarity between trials using image hashing"""
        trials = [self._config["trial_1_name"], self._config["trial_2_name"]]
        cameras = ["cam1", "cam2"]
        videos = {
            (trial, cam): cv2.VideoCapture(
                self._data_processor.find_cam_file(self._trials_path / trial, cam)
            )
            for trial in trials
            for cam in cameras
        }

        # Compare hashes based on the camera views configuration
        if self._config["cam1s_are_the_same_view"]:
            cam1_diff, noc1 = self._compare_two_videos(
                videos[(trials[0], "cam1")], videos[(trials[1], "cam1")]
            )
            cam2_diff, noc2 = self._compare_two_videos(
                videos[(trials[0], "cam2")], videos[(trials[1], "cam2")]
            )
        else:
            cam1_diff, noc1 = self._compare_two_videos(
                videos[(trials[0], "cam1")], videos[(trials[1], "cam2")]
            )
            cam2_diff, noc2 = self._compare_two_videos(
                videos[(trials[0], "cam2")], videos[(trials[1], "cam1")]
            )

        # Calculate the average difference for each camera view
        cam1_avg_diff = cam1_diff / noc1 if noc1 > 0 else 0
        cam2_avg_diff = cam2_diff / noc2 if noc2 > 0 else 0

        # Calculate the overall trial average difference
        trial_avg_diff = (cam1_avg_diff + cam2_avg_diff) / 2

        # Note: The number of comparisons (noc) grows with video size, which could potentially affect the similarity measure,
        # making larger videos appear more similar than they actually are. This aspect may need further consideration.
        return trial_avg_diff

    def analyze_marker_similarity_project(self):
        """Analyze all videos in a project and get their average rhythmicity. This assumes that all cam1/2 pairs are either the same or different!"""
        marker_similarity = {}

        trial_perms = combinations(self._data_processor.list_trials(), 2)
        logger.debug(f"Trial permutations for project: {trial_perms}")
        for trial1, trial2 in trial_perms:
            self._config["trial_1_name"] = trial1.stem
            self._config["trial_2_name"] = trial2.stem
            with self.project_config.open("w") as file:
                yaml.dump(self._config, file, sort_keys=False)
            # TODO: pass the trial names directly instead of writing into config
            marker_similarity[(trial1, trial2)] = abs(
                self.analyze_marker_similarity_trial()
            )
        return marker_similarity

    def analyze_marker_similarity_trial(self):
        """Analyze marker similarity for a pair of trials using the distance formula."""
        # Find CSVs for each trial
        trial1 = self._config["trial_1_name"]
        trial2 = self._config["trial_2_name"]
        trial1_path = self._trials_path / trial1
        current_files = trial1_path.glob("*")
        logger.debug(f"Current files in directory {trial1_path}: {current_files}")
        trial2_path = self._trials_path / trial2
        current_files = trial2_path.glob("*")
        logger.debug(f"Current files in directory {trial2_path}: {current_files}")

        # Get a list of markers that each trial have in common
        bodyparts1_csv_path = self._data_processor.find_trial_csv(trial1_path)
        bodyparts2_csv_path = self._data_processor.find_trial_csv(trial2_path)
        bodyparts1 = self._data_processor.get_bodyparts_from_xma(
            bodyparts1_csv_path, mode="rgb"
        )
        bodyparts2 = self._data_processor.get_bodyparts_from_xma(
            bodyparts2_csv_path, mode="rgb"
        )
        markers_in_common = [marker for marker in bodyparts1 if marker in bodyparts2]
        logger.debug(f"Markers in common for similarity analysis: {markers_in_common}")

        # Analyze intermarker distances for each marker in common
        trial1_csv = pd.read_csv(bodyparts1_csv_path)
        trial2_csv = pd.read_csv(bodyparts2_csv_path)
        avg_distances = []
        for marker in markers_in_common:
            # Debug logging
            x1_vals = trial1_csv[f"{marker}_X"]
            logger.debug(f"Trial1 values for {marker}_X: {x1_vals}")
            x2_vals = trial2_csv[f"{marker}_X"]
            logger.debug(f"Trial2 values for {marker}_X: {x2_vals}")

            # Get mean positions for each marker
            avg_x1, avg_y1 = (
                trial1_csv[f"{marker}_X"].mean(),
                trial1_csv[f"{marker}_Y"].mean(),
            )
            logger.debug(
                f"Average trial1 position for marker {marker}: ({avg_x1}, {avg_y1})"
            )
            avg_x2, avg_y2 = (
                trial2_csv[f"{marker}_X"].mean(),
                trial2_csv[f"{marker}_Y"].mean(),
            )
            logger.debug(
                f"Average trial2 position for marker {marker}: ({avg_x2}, {avg_y2})"
            )

            # Calculate the distance between the average positions for this marker in the two trials
            distance = math.sqrt((avg_x2 - avg_x1) ** 2 + (avg_y2 - avg_y1) ** 2)
            avg_distances.append(distance)

        logger.debug(f"Avg distances for trials {trial1} and {trial2}: {avg_distances}")
        # Calculate the mean of the distances to get an overall similarity measure
        marker_similarity = (
            sum(avg_distances) / len(avg_distances) if avg_distances else 0
        )

        return marker_similarity

    def get_max_dissimilarity_for_trial(self, trial_path, window):
        """Calculate the dissimilarity within the trial given the frame sliding window."""
        video1 = cv2.VideoCapture(
            self._data_processor.find_cam_file(trial_path, "cam1")
        )
        video2 = cv2.VideoCapture(
            self._data_processor.find_cam_file(trial_path, "cam2")
        )

        hashes1 = self._hash_trial_video(video1)
        hashes2 = self._hash_trial_video(video2)

        return self._find_dissimilar_regions(hashes1, hashes2, window)

    def _compare_two_videos(self, video1, video2):
        """Compare two videos using image hashing"""
        hashes1 = self._hash_trial_video(video1)
        hashes2 = self._hash_trial_video(video2)

        video1_frames = len(hashes1)
        video2_frames = len(hashes2)
        # Thought: maybe this should be computed alongside the for loop?
        noc = math.perm(
            video1_frames + video2_frames, 2
        )  # Might need revision based on actual comparison logic

        print(f"Video 1 frames: {video1_frames}")
        print(f"Video 2 frames: {video2_frames}")

        print("Comparing hashes between videos")
        hash_dif = sum(hash1 - hash2 for hash1 in hashes1 for hash2 in hashes2)

        return hash_dif, noc

    def _hash_trial_video(self, video):
        """Generate image hashes for a single video"""
        video_frames = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
        print(f"Video frames: {video_frames}")

        hashes = []
        print("Creating hashes for video")
        for i in range(video_frames):
            print(f"Current frame: {i}")
            ret, frame = video.read()
            if not ret:
                # Should this throw an error?
                cv2.destroyAllWindows()
                raise IOError("Error reading video frame")
            hashes.append(imagehash.phash(Image.fromarray(frame)))

        return hashes

    def _find_dissimilar_regions(self, hashes1, hashes2, window):
        """Find the region of maximum dissimilarity given 2 lists of hashes and a sliding window (how many frames)"""
        start_frame_vid1 = 0
        start_frame_vid2 = 0
        max_hash_dif_vid1 = 0
        max_hash_dif_vid2 = 0
        hash_dif_vid1 = 0
        hash_dif_vid2 = 0

        for slider in range(0, len(hashes1) // window):
            print(f"Current start frame {slider * window}")
            hash_dif_vid1, hash_dif_vid2 = self._compare_hash_sets(
                hashes1[slider * window : (slider + 1) * window],
                hashes2[slider * window : (slider + 1) * window],
            )

            print(f"Current hash diff (vid 1): {hash_dif_vid1}")
            print(f"Current hash diff (vid 2): {hash_dif_vid2}")
            if hash_dif_vid1 > max_hash_dif_vid1:
                max_hash_dif_vid1 = hash_dif_vid1
                start_frame_vid1 = slider * window

            if hash_dif_vid2 > max_hash_dif_vid2:
                max_hash_dif_vid2 = hash_dif_vid2
                start_frame_vid2 = slider * window

            print(f"Max hash diff (vid 1): {max_hash_dif_vid1}")
            print(f"Max hash diff (vid 2): {max_hash_dif_vid2}")

            print(f"Start frame (vid 1): {start_frame_vid1}")
            print(f"Start frame (vid 2): {start_frame_vid2}")

        return start_frame_vid1, start_frame_vid2

    def _compare_hash_sets(self, hashes1, hashes2):
        """Compares two sets of image hashes to find dissimilarities"""
        hash1_dif = 0
        hash2_dif = 0

        print(f"Hash set 1 {hashes1[0]}")
        print(f"Hash set 2 {hashes2[0]}")
        # Compares all possible combinations of images
        for combination in combinations(hashes1, 2):
            hash1_dif = hash1_dif + (combination[0] - combination[1])

        for combination in combinations(hashes2, 2):
            hash2_dif = hash2_dif + (combination[0] - combination[1])

        return hash1_dif, hash2_dif



"""
XROMM Tools for DeepLabCut
Developed by J.D. Laurence-Chasen

Functions:

xma_to_dlc: create DeepLabCut training dataset from data tracked in XMALab
analyze_xromm_videos: Predict 2D points for novel trials
dlc_to_xma: convert output of DeepLabCut to XMALab format 2D points file
add_frames: Add new frames corrected/tracked in XMALab to an existing training dataset

"""

from pathlib import Path
import pandas as pd
import numpy as np
import cv2


def dlc_to_xma(trial: Path, iteration: int):
    # get filenames and read analyzed data
    trialname = trial.name
    iteration_folder = trial / f"it{iteration}"
    cam1data = list(iteration_folder.glob("*cam1*.h5"))
    cam2data = list(iteration_folder.glob("*cam2*.h5"))
    if not cam1data:
        raise ValueError(
            "Cannot find cam1 predicted points. Have you run deepxromm.analyze_videos() for this project?"
        )
    if not cam2data:
        raise ValueError(
            "Cannot find cam2 predicted points. Have you run deepxromm.analyze_videos() for this project?"
        )
    cam1data = pd.read_hdf(cam1data[0])
    cam2data = pd.read_hdf(cam2data[0])
    h5_save_path = iteration_folder / f"{trialname}-Predicted2DPoints.h5"
    csv_save_path = iteration_folder / f"{trialname}-Predicted2DPoints.csv"

    pointnames = list(cam1data.columns.get_level_values("bodyparts").unique())

    # make new column names
    nvar = len(pointnames)
    pointnames = [item for item in pointnames for _ in range(4)]
    post = ["_cam1_X", "_cam1_Y", "_cam2_X", "_cam2_Y"] * nvar
    cols = [m + str(n) for m, n in zip(pointnames, post)]

    # remove likelihood columns
    cam1data = cam1data.drop(cam1data.columns[2::3], axis=1)
    cam2data = cam2data.drop(cam2data.columns[2::3], axis=1)

    # replace col names with new indices
    c1cols = list(range(0, cam1data.shape[1] * 2, 4)) + list(
        range(1, cam1data.shape[1] * 2, 4)
    )
    c2cols = list(range(2, cam1data.shape[1] * 2, 4)) + list(
        range(3, cam1data.shape[1] * 2, 4)
    )
    c1cols.sort()
    c2cols.sort()
    cam1data.columns = c1cols
    cam2data.columns = c2cols

    df = pd.concat([cam1data, cam2data], axis=1).sort_index(axis=1)
    df.columns = cols
    df.to_hdf(h5_save_path, key="df_with_missing", mode="w")
    df.to_csv(csv_save_path, na_rep="NaN", index=False)


def add_frames(
    path_config_file,
    data_path,
    iteration,
    frames,
    nnetworks=1,
    path_config_file_cam2="enterpathofcam2config",
):
    # input: config file paths, path of data to add to trainingdataset, frames-csv file where first col is trialnames and following cols are frame numbers
    # will look for 2D points file based on name (if there are multiple csv files)

    configs = [path_config_file[:-12], path_config_file_cam2[:-12]]
    cameras = [1, 2]
    subs = [
        [
            "c01",
            "c1",
            "C01",
            "C1",
            "Cam1",
            "cam1",
            "Cam01",
            "cam01",
            "Camera1",
            "camera1",
        ],
        [
            "c02",
            "c2",
            "C02",
            "C2",
            "Cam2",
            "cam2",
            "Cam02",
            "cam02",
            "Camera2",
            "camera2",
        ],
    ]
    pts = [
        "2Dpts",
        "2dpts",
        "2DPts",
        "2dPts",
        "pts2D",
        "Pts2D",
        "pts2d",
        "points2D",
        "Points2d",
        "points2d",
        "2Dpoints",
        "2dpoints",
        "2DPoints",
    ]
    corr = ["correct", "Correct", "corrected", "Corrected"]
    # read frames from csv
    if ".csv" in frames:
        f = pd.read_csv(frames, header=None)
        trialnames = list(f.iloc[:, 0])  # first row of frames file must be trialnames
        picked_frames = []
        # this is disgusting code
        for row in range(f.shape[0]):
            picked_frames.append(list(f.loc[row, 1:]))
        for count, row in enumerate(picked_frames):
            picked_frames[count] = [x for x in row if str(x) != "nan"]  # remove nans
        for count, row in enumerate(picked_frames):
            picked_frames[count] = [int(x) for x in row]  # convert to int
    else:
        raise ValueError("frames must be a .csv file with trialnames and frame numbers")

    if nnetworks == 2:
        for camera in cameras:
            dlc_dataset_path = configs[camera - 1] / "labeled-data"
            contents = list(dlc_dataset_path.glob("*"))
            if len(contents) == 1:
                dataset_name = contents[0]
                labeleddata_path = dlc_dataset_path / dataset_name
            else:
                raise ValueError(
                    "There must be only one data set in the labeled-data folder"
                )

            h5file = list(labeleddata_path.glob("*.h5"))
            csvfile = list(labeleddata_path.glob("*.csv"))
            data = pd.read_hdf(labeleddata_path / h5file[0])

            ## Extract selected frames from videos

            for trialnum, trial in enumerate(trialnames):
                # get video file
                file = []
                relnames = []
                contents = data_path / trial
                for name in contents:
                    if any(x in name for x in subs[camera - 1]):
                        file = name
                if not file:
                    raise ValueError(
                        "Cannot locate %s video file or image folder" % trial
                    )

                # if video file is actually folder of frames
                video_path = data_path / trial / file
                if video_path.is_dir():
                    imgpath = video_path
                    imgs = imgpath.glob("*")
                    relpath = Path("labeled-data" / dataset_name)
                    frames = picked_frames[trialnum]
                    frames.sort()

                    for count, img in enumerate(imgs):
                        if count + 1 in frames:  # ASSUMES FRAMES PROVIDED ARE 1 index
                            image = cv2.imread(imgpath / img)
                            current_img = str(count + 1).zfill(4)
                            relname = relpath / f"{trial}_{current_img}.png"
                            relnames = relnames + [relname]
                            cv2.imwrite(
                                labeleddata_path / trial / f"{trial}_{current_img}.png",
                                image,
                            )
                else:
                    relpath = Path("labeled-data" / dataset_name)
                    frames = picked_frames[trialnum]
                    frames.sort()
                    cap = cv2.VideoCapture(video_path)
                    success, image = cap.read()
                    count = 0
                    while success:
                        if count + 1 in frames:
                            current_img = str(count + 1).zfill(4)
                            relname = relpath / f"{trial}_{current_img}.png"
                            relnames = relnames + [relname]
                            cv2.imwrite(
                                labeleddata_path / trial / f"{trial}_{current_img}.png",
                                image,
                            )
                        success, image = cap.read()
                        count += 1
                    cap.release()

                # get 2D points file / data
                # extract 2D points data
                iteration_dir = data_path / trial / f"it{iteration}"
                pointsfile = list(iteration_dir.glob("*.csv"))

                if not pointsfile:
                    raise ValueError(f"Cannot locate {trial} 2D points file")

                # if multiple csv files, look for "2Dpoints" in the name
                if len(pointsfile) > 1:
                    t = []
                    for q in pointsfile:
                        if any(x in q for x in pts):
                            t = t + [q]
                    # if there are multiple 2D points files, look for "corrected" in the name
                    if len(t) > 1:
                        for r in pointsfile:
                            if any(x in r for x in corr):
                                file = r
                    else:
                        file = t[0]
                else:
                    file = pointsfile[0]
                print(
                    f"Reading and adding the following frames from {iteration_dir}/{file}"
                )
                df = pd.read_csv(iteration_dir / file, sep=",", header=None)
                df = df.loc[1:,].reset_index(drop=True)
                print(frames)
                frames = [x - 1 for x in frames]  # account for zero index in python
                xpos = df.iloc[frames, 0 + (camera - 1) * 2 :: 4]
                ypos = df.iloc[frames, 1 + (camera - 1) * 2 :: 4]
                temp_data = pd.concat([xpos, ypos], axis=1).sort_index(axis=1)
                if temp_data.shape[1] > data.shape[1]:
                    raise ValueError(
                        "There are %d extra points in the corrected points file"
                        % ((temp_data.shape[1] - data.shape[1]) / 2)
                    )
                if temp_data.shape[1] < data.shape[1]:
                    raise ValueError(
                        "There are %d missing points in the corrected points file"
                        % ((data.shape[1] - temp_data.shape[1]) / 2)
                    )
                temp_data.index = relnames
                temp_data.columns = data.columns
                data = pd.concat([data, temp_data])
            data.replace(" NaN", np.nan, inplace=True)
            data.replace(" NaN ", np.nan, inplace=True)
            data.replace("NaN ", np.nan, inplace=True)
            data = data.astype("float")
            data = data.round(2)
            data = data.apply(pd.to_numeric)
            data.to_hdf(
                labeleddata_path + "/" + h5file[0], key="df_with_missing", mode="w"
            )
            data.to_csv(labeleddata_path + "/" + csvfile[0], na_rep="NaN")

    else:  # default, one network for both videos
        labeled_data_dir = config / "labeled-data"
        print(str(labeled_data_dir))
        contents = list(labeled_data_dir.glob("*"))
        if len(contents) == 1:
            dataset_name = contents[0]
            labeleddata_path = labeled_data_dir / dataset_name
        else:
            raise ValueError(
                "There must be only one data set in the labeled-data folder"
            )

        h5file = list(labeleddata_path.glob("*.h5"))
        csvfile = list(labeleddata_path.glob("*.csv"))
        data = pd.read_hdf(labeleddata_path / h5file[0])  # read old point labels

        for camera in cameras:
            ## Extract selected frames from videos
            for trialnum, trial in enumerate(trialnames):
                # get video file
                relnames = []
                file = []

                trial_path = data_path / trial
                contents = trial_path.glob("*")
                for name in contents:
                    if any(x in name for x in subs[camera - 1]):
                        file = name
                if not file:
                    raise ValueError(
                        "Cannot locate %s video file or image folder" % trial
                    )

                # if video file is actually folder of frames
                video_path = data_path / trial / file
                if video_path.is_dir():
                    imgpath = video_path
                    imgs = imgpath.glob("*")
                    relpath = Path("labeled-data" / dataset_name)
                    frames = picked_frames[trialnum]
                    frames.sort()

                    for count, img in enumerate(imgs):
                        if count + 1 in frames:  # ASSUMES FRAMES PROVIDED ARE 1 index
                            image = cv2.imread(imgpath / img)
                            current_img = str(count + 1).zfill(4)
                            relname = relpath / f"{trial}_cam{camera}_{current_img}.png"
                            relnames = relnames + [relname]
                            cv2.imwrite(
                                labeleddata_path
                                / f"{trial}_cam{camera}_{current_img}.png"
                            )
                else:
                    # file is actually a file
                    # extract frames from video and convert to png
                    relpath = Path("labeled-data" / dataset_name)
                    frames = picked_frames[trialnum]
                    frames.sort()
                    cap = cv2.VideoCapture(video_path)
                    success, image = cap.read()
                    count = 0
                    while success:
                        if count + 1 in frames:
                            current_img = str(count + 1).zfill(4)
                            relname = relpath / f"{trial}_cam{camera}_{current_img}.png"
                            relnames = relnames + [relname]
                            cv2.imwrite(
                                labeleddata_path
                                / f"{trial}_cam{camera}_{current_img}.png"
                            )
                        success, image = cap.read()
                        count += 1
                    cap.release()

                # get 2D points file / data
                # extract 2D points data
                contents = data_path / trial / f"it{iteration}"
                pointsfile = list(contents.glob("*.csv"))

                if not pointsfile:
                    raise ValueError(f"Cannot locate {trial} 2D points file")

                # if multiple csv files, look for "2Dpoints" in the name
                if len(pointsfile) > 1:
                    t = []
                    for q in pointsfile:
                        if any(x in q for x in pts):
                            t = t + [q]
                    # if there are multiple 2D points files, look for "corrected" in the name
                    if len(t) > 1:
                        for r in pointsfile:
                            if any(x in r for x in corr):
                                pointsfile = r
                else:
                    pointsfile = pointsfile[0]
                if isinstance(pointsfile, str) != True:
                    raise ValueError(
                        f"Please check the points files in trial {trial} iteration {iteration} folder"
                    )
                df = pd.read_csv(
                    data_path / trial / f"it{iteration}" / pointsfile,
                    sep=",",
                    header=None,
                )
                df = df.loc[1:,].reset_index(drop=True)
                xpos = df.iloc[frames, 0 + (camera - 1) * 2 :: 4]
                ypos = df.iloc[frames, 1 + (camera - 1) * 2 :: 4]
                temp_data = pd.concat([xpos, ypos], axis=1).sort_index(axis=1)
                temp_data.index = relnames
                if temp_data.shape[1] > data.shape[1]:
                    raise ValueError(
                        "There are %d extra points in the corrected points file"
                        % ((temp_data.shape[1] - data.shape[1]) / 2)
                    )
                if temp_data.shape[1] < data.shape[1]:
                    raise ValueError(
                        "There are %d missing points in the corrected points file"
                        % ((data.shape[1] - temp_data.shape[1]) / 2)
                    )
                temp_data.columns = data.columns
                data = pd.concat([data, temp_data])
        data.replace(" NaN", np.nan, inplace=True)
        data.replace(" NaN ", np.nan, inplace=True)
        data.replace("NaN ", np.nan, inplace=True)
        data = data.astype("float")
        data = data.round(2)
        data = data.apply(pd.to_numeric)
        data.to_hdf(labeleddata_path / h5file[0], key="df_with_missing", mode="w")
        data.to_csv(labeleddata_path / csvfile[0], na_rep="NaN")

    print(
        f"Frames from {len(trialnames)} trials successfully added to training dataset"
    )



"""Primary interface for training the XROMM network using DLC"""

from pathlib import Path

import deeplabcut
import pandas as pd

from .xma_data_processor import XMADataProcessor


class Network:
    """Trains an XROMM labeling network using DLC."""

    def __init__(self, config):
        self.working_dir = Path(config["working_dir"])
        self._trainingdata_path = self.working_dir / "trainingdata"  # Keep for RGB mode
        self._data_processor = XMADataProcessor(config)
        self._config = config

    def xma_to_dlc(self) -> None:
        """Convert XMAlab data to DLC format"""
        mode = self._config["mode"]
        trials = self._data_processor.list_trials("trainingdata")
        dfs, idx, pointnames = self._data_processor.read_trial_csv_with_validation(
            trials
        )

        # Validate we have enough frames
        total_frames = sum(len(x) for x in idx)
        nframes = int(self._config["nframes"])
        if total_frames < nframes:
            raise ValueError(
                f"Requested {nframes} frames but only found {total_frames} "
                f"valid frames across {len(trials)} trials"
            )

        picked_frames = self._data_processor._extract_frame_selection_loop(idx, nframes)

        cameras = [1, 2]
        if mode == "2D":
            self._process_cameras_2d(trials, picked_frames, dfs, pointnames, cameras)

        elif mode == "per_cam":
            config_files = [
                Path(self._config["path_config_file"]).parent,
                Path(self._config["path_config_file_2"]).parent,
            ]
            for camera, config_file in zip(cameras, config_files):
                self._process_camera_per_cam(
                    camera, config_file, trials, picked_frames, dfs, pointnames
                )

        elif mode == "rgb":
            print("We've selected an RGB video")
            self._data_processor.make_rgb_videos("trainingdata")
            self._data_processor.xma_to_dlc_rgb("trainingdata")
        else:
            raise AttributeError(f"Unsupportede mode: {mode}")

    def create_training_dataset(self):
        """Create training dataset for data"""
        deeplabcut.create_training_dataset(self._config["path_config_file"])
        if self._config["mode"] == "per_cam":
            deeplabcut.create_training_dataset(self._config["path_config_file_2"])

    def train(self):
        """Starts training a network"""
        deeplabcut.train_network(
            self._config["path_config_file"], maxiters=self._config["maxiters"]
        )

        if self._config["mode"] == "per_cam":
            deeplabcut.train_network(
                self._config["path_config_file_2"], maxiters=self._config["maxiters"]
            )

    def _process_cameras_2d(
        self,
        trials: list[Path],
        picked_frames: list[list[int]],
        dfs: list[pd.DataFrame],
        pointnames: list[str],
        cameras: list[int] = [1, 2],
    ):
        """
        Process both cameras for 2D mode (nnetworks=1).

        Both cameras are combined into a single DLC project and dataset.

        Args:
            trials: List of trial directory paths
            picked_frames: List of frame indices per trial
            dfs: List of DataFrames (one per trial)
            pointnames: List of body part names
            cameras: List of cameras
        """

        config_dir = Path(self._config["path_config_file"]).parent
        dataset_name = self._config["dataset_name"]
        newpath = config_dir / "labeled-data" / dataset_name
        if newpath.exists():
            contents = list(newpath.glob("*"))
            if len(contents) > 0:
                print(
                    f"Directory {newpath} already contains data. "
                    "Please use a different dataset name or clear the directory."
                )
                return
        else:
            newpath.mkdir(parents=True, exist_ok=True)

        relnames = []
        data = pd.DataFrame()

        for camera in cameras:
            print(f"Extracting camera {camera} trial images and 2D points...")
            for trialnum, trial in enumerate(trials):
                # Find the camera video/image source
                cam_identifier = f"cam{camera}"
                source_path = self._data_processor.find_cam_file(trial, cam_identifier)

                # Sort frames to extract for given trial
                frames = sorted(picked_frames[trialnum])
                trial_relnames = self._data_processor.extract_frames_from_video(
                    source_path=source_path,
                    frame_indices=frames,
                    output_dir=newpath,
                    output_name_base=trial.name,
                    mode="2D",
                    camera=camera,
                    compression=0,
                )
                relnames.extend(trial_relnames)

                # Extract 2D points for this camera
                temp_data = self._data_processor.extract_2d_points_for_camera(
                    dfs[trialnum], camera, frames
                )

                # Reset column names for combining data from multiple cameras
                temp_data.columns = range(temp_data.shape[1])
                data = pd.concat([data, temp_data])

        # Create and save DLC dataset
        self._data_processor.save_dlc_dataset(
            data, self._config["experimenter"], relnames, pointnames, newpath
        )
        print("DLC dataset extracted from provided XMAlab trials")

    def _process_camera_per_cam(
        self,
        camera: int,
        config_dir: Path,
        trialnames: list[Path],
        picked_frames: list[list[int]],
        dfs: list[pd.DataFrame],
        pointnames: list[str],
    ) -> None:
        """
        Process single camera for per_cam mode (nnetworks=2).

        Each camera gets its own separate DLC project and dataset.

        Args:
            camera: Camera number (1 or 2)
            config_dir: DLC config file parent directory for this camera
            trialnames: List of trial directory paths
            picked_frames: List of frame indices per trial
            dfs: List of DataFrames (one per trial)
            pointnames: List of body part names
            data_processor: XMADataProcessor instance
        """
        print(f"Extracting camera {camera} trial images and 2D points...")

        # Setup output directory with camera-specific dataset name
        dataset_name = self._config["dataset_name"]
        scorer = self._config["experimenter"]
        camera_dataset_name = f"{dataset_name}_cam{camera}"
        newpath = config_dir / "labeled-data" / camera_dataset_name
        if newpath.exists():
            contents = list(newpath.glob("*"))
            if len(contents) > 0:
                print(
                    f"Directory {newpath} already contains data. "
                    "Please use a different dataset name or remove the existing dataset to update it"
                )
                return
        else:
            newpath.mkdir(parents=True, exist_ok=True)

        # Process each trial
        relnames = []
        data = pd.DataFrame()

        for trialnum, trial_path in enumerate(trialnames):
            trial_name = trial_path.name

            # Extract frames using unified interface
            frames = sorted(picked_frames[trialnum])
            # Find the camera video/image source
            cam_identifier = f"cam{camera}"
            cam_file = self._data_processor.find_cam_file(trial_path, cam_identifier)
            source_path = trial_path / cam_file

            trial_relnames = self._data_processor.extract_frames_from_video(
                source_path=source_path,
                frame_indices=frames,
                output_dir=newpath,
                output_name_base=trial_name,
                mode="per_cam",
                camera=camera,
                compression=0,
            )
            relnames.extend(trial_relnames)

            # Extract 2D points for this camera
            temp_data = self._data_processor.extract_2d_points_for_camera(
                dfs[trialnum], camera, frames
            )
            data = pd.concat([data, temp_data])

        # Create and save DLC dataset
        self._data_processor.save_dlc_dataset(
            data, scorer, relnames, pointnames, newpath
        )
        print("DLC dataset extracted from provided XMAlab trials")



"""
Unit tests for deepxromm project creation
"""

import os
from pathlib import Path
import shutil
import unittest

import yaml

from deepxromm import DeepXROMM

SAMPLE_FRAME = Path(__file__).parent / "sample_frame.jpg"
SAMPLE_FRAME_INPUT = Path(__file__).parent / "sample_frame_input.csv"
SAMPLE_AUTOCORRECT_OUTPUT = Path(__file__).parent / "sample_autocorrect_output.csv"

DEEPXROMM_TEST_CODEC = os.environ.get("DEEPXROMM_TEST_CODEC", "avc1")


class TestProjectCreation(unittest.TestCase):
    """Tests behaviors related to XMA-DLC project creation"""

    @classmethod
    def setUpClass(cls):
        """Create a sample project"""
        super(TestProjectCreation, cls).setUpClass()
        cls.project_dir = Path.cwd() / "tmp"
        DeepXROMM.create_new_project(cls.project_dir, codec=DEEPXROMM_TEST_CODEC)

    def test_project_creates_correct_folders(self):
        """Do we have all of the correct folders?"""
        for folder in ["trainingdata", "trials"]:
            with self.subTest(folder=folder):
                self.assertTrue((self.project_dir / folder).exists())

    def test_project_creates_config_file(self):
        """Do we have a project config?"""
        self.assertTrue((self.project_dir / "project_config.yaml").exists())

    def test_project_config_has_these_variables(self):
        """Can we access each of the variables that's supposed to be in the config?"""
        variables = [
            "task",
            "experimenter",
            "working_dir",
            "path_config_file",
            "dataset_name",
            "nframes",
            "maxiters",
            "tracking_threshold",
            "mode",
            "swapped_markers",
            "crossed_markers",
            "search_area",
            "threshold",
            "krad",
            "gsigma",
            "img_wt",
            "blur_wt",
            "gamma",
            "cam",
            "frame_num",
            "trial_name",
            "marker",
            "test_autocorrect",
            "cam1s_are_the_same_view",
        ]

        config_path = self.project_dir / "project_config.yaml"
        with config_path.open() as config:
            project = yaml.safe_load(config)
            for variable in variables:
                with self.subTest(i=variable):
                    self.assertIsNotNone(project[variable])

    @classmethod
    def tearDownClass(cls):
        """Remove the created temp project"""
        super(TestProjectCreation, cls).tearDownClass()
        shutil.rmtree(cls.project_dir)



"""
Test of the 'train_many_projects' batch training pipeline
"""

import unittest
import os
from pathlib import Path
import yaml
import shutil

from deepxromm import DeepXROMM

DEEPXROMM_TEST_CODEC = os.environ.get("DEEPXROMM_TEST_CODEC", "avc1")


def set_up_project(project_dir: Path, mode: str):
    """Helper function to set up a project at a given path with a given mode"""
    deepxromm_proj = DeepXROMM.create_new_project(
        project_dir, mode=mode, codec=DEEPXROMM_TEST_CODEC
    )

    # Adjust maxiters to 5 to ensure that training completes quickly
    deepxromm_proj.config["maxiters"] = 5
    print(deepxromm_proj.config)
    with (project_dir / "project_config.yaml").open("w") as fp:
        yaml.dump(deepxromm_proj.config, fp, sort_keys=False)

    # Make vars for pathing to find files easily
    trial_dir = project_dir / "trainingdata/test"
    trial_dir.mkdir(parents=True, exist_ok=True)
    trial_csv = trial_dir / "test.csv"
    cam1_path = trial_dir / "test_cam1.avi"
    cam2_path = trial_dir / "test_cam2.avi"

    # Copy sample CSV data for training operations (use existing sample file)
    shutil.copy("trial_slice.csv", str(trial_csv))
    shutil.copy("trial_cam1_slice.avi", str(cam1_path))
    shutil.copy("trial_cam2_slice.avi", str(cam2_path))

    # Create sample novel trial for analysis
    novel_dir = project_dir / "trials/test"
    novel_dir.mkdir(parents=True, exist_ok=True)

    # Copy videos for analysis
    shutil.copy("trial_cam1_slice.avi", str(novel_dir / "test_cam1.avi"))
    shutil.copy("trial_cam2_slice.avi", str(novel_dir / "test_cam2.avi"))


class TestBatchTrainer(unittest.TestCase):
    def setUp(self):
        """Create trial with test data"""
        self.working_dir = Path.cwd() / "tmp"
        # Create 2D base project
        self.base_project_dir = self.working_dir / "2D"
        set_up_project(self.base_project_dir, "2D")

        # Create per_cam project
        self.per_cam_project_dir = self.working_dir / "per_cam"
        set_up_project(self.per_cam_project_dir, "per_cam")

        # Create RGB project
        self.rgb_project_dir = self.working_dir / "rgb"
        set_up_project(self.rgb_project_dir, "rgb")

    def test_train_many_projects(self):
        """Invoke train_many_projects"""
        # Invoke once to run through the whole process for each project type
        DeepXROMM.train_many_projects(self.working_dir)

        # Invoke again to ensure batch workflows complete idempotently (i.e. skips are not fatal)
        DeepXROMM.train_many_projects(self.working_dir)

    def tearDown(self):
        """Remove the created temp project"""
        if self.working_dir.exists():
            shutil.rmtree(self.working_dir)



"""A Complete Set of User-Friendly Tools for DeepLabCut-XMAlab marker tracking"""

from pathlib import Path
from ruamel.yaml.comments import CommentedMap

from .analyzer import Analyzer
from .autocorrector import Autocorrector
from .network import Network
from .project import Project
from .xma_data_processor import XMADataProcessor


class DeepXROMM:
    """A Complete Set of User-Friendly Tools for DeepLabCut-XMAlab marker tracking"""

    config: CommentedMap
    _analyzer: Analyzer
    _autocorrector: Autocorrector
    _network: Network
    _data_processor: XMADataProcessor

    def __init__(self):
        # Prevent direct instantiation because we need to do different
        # initialization work in create_ vs load_ methods.
        raise NotImplementedError("Use create_new_project or load_project instead.")

    @classmethod
    def create_new_project(
        cls, working_dir=None, experimenter="NA", mode="2D", codec="avc1"
    ):
        """Create a new xrommtools project"""
        deepxromm = DeepXROMM.__new__(DeepXROMM)
        deepxromm.config = Project.create_new_config(
            working_dir, experimenter, mode, codec
        )
        deepxromm._analyzer = Analyzer(deepxromm.config)
        deepxromm._autocorrector = Autocorrector(deepxromm.config)
        deepxromm._network = Network(deepxromm.config)
        deepxromm._data_processor = XMADataProcessor(deepxromm.config)
        return deepxromm

    @classmethod
    def load_project(cls, working_dir=None):
        """Create a new xrommtools project"""
        deepxromm = DeepXROMM.__new__(DeepXROMM)
        deepxromm.config = Project.load_config(working_dir)
        deepxromm._analyzer = Analyzer(deepxromm.config)
        deepxromm._autocorrector = Autocorrector(deepxromm.config)
        deepxromm._network = Network(deepxromm.config)
        deepxromm._data_processor = XMADataProcessor(deepxromm.config)
        return deepxromm

    def xma_to_dlc(self):
        """Converts XMAlab-compatible data to DLC format"""
        self._network.xma_to_dlc()

    def create_training_dataset(self):
        """Creates a training dataset based on current project data"""
        self._network.create_training_dataset()

    def train_network(self):
        """Starts training the network using data in the working directory."""
        self._network.train()

    def analyze_videos(self):
        """Analyze videos with a pre-existing network"""
        self._analyzer.analyze_videos()

    def dlc_to_xma(self):
        self._data_processor.dlc_to_xma()

    def autocorrect_trials(self):
        """Do XMAlab-style autocorrect on the tracked beads for all trials"""
        self._autocorrector.autocorrect_trials()

    def get_bodyparts_from_xma(self, csv_path: str, mode: str):
        """Pull the names of the XMAlab markers from the 2Dpoints file"""
        return self._data_processor.get_bodyparts_from_xma(csv_path, mode)

    def split_rgb(self, trial_path, codec=None):
        """Takes a RGB video with different grayscale data written to the R, G,
        and B channels and splits it back into its component source videos.

        Args:
            trial_path: Path to trial directory containing RGB video
            codec: Video codec to use. If None, uses video_codec from config.
                   Common options: "avc1", "DIVX", "XVID", "mp4v", "MJPG", "uncompressed"

        Raises:
            RuntimeError: If the specified codec is not available on this system
        """
        return self._data_processor.split_rgb(trial_path, codec)

    def analyze_video_similarity_project(self):
        """Analyze all videos in a project and take their average similar. This
        is dangerous, as it will assume that all cam1/cam2 pairs match or don't match!
        """
        return self._analyzer.analyze_video_similarity_project()

    def analyze_video_similarity_trial(self):
        """Analyze the average similarity between trials using image hashing"""
        return self._analyzer.analyze_video_similarity_trial()

    def get_max_dissimilarity_for_trial(self, trial_path, window):
        """Calculate the dissimilarity within the trial given the frame sliding window."""
        return self._analyzer.get_max_dissimilarity_for_trial(trial_path, window)

    def analyze_marker_similarity_project(self):
        """Analyze all videos in a project and get their average rhythmicity. This assumes
        that all cam1/2 pairs are either the same or different!"""
        return self._analyzer.analyze_marker_similarity_project()

    def analyze_marker_similarity_trial(self):
        """Analyze marker similarity for a pair of trials. Returns the mean difference for
        paired marker positions (X - X, Y - Y for each marker)"""
        return self._analyzer.analyze_marker_similarity_trial()

    @staticmethod
    def train_many_projects(parent_dir):
        """Train and analyze multiple deepxromm projects given a parent folder"""
        parent_path = Path(parent_dir)
        for folder in parent_path.iterdir():
            if not folder.is_dir():
                continue
            project_path = folder
            deepxromm = DeepXROMM.load_project(str(project_path))
            deepxromm.xma_to_dlc()
            deepxromm.create_training_dataset()
            deepxromm.train_network()
            deepxromm.analyze_videos()
            deepxromm.dlc_to_xma()



import os
from pathlib import Path
import shutil
import unittest
import random

import pandas as pd

from deepxromm import DeepXROMM

SAMPLE_FRAME = Path(__file__).parent / "sample_frame.jpg"
SAMPLE_FRAME_INPUT = Path(__file__).parent / "sample_frame_input.csv"
SAMPLE_AUTOCORRECT_OUTPUT = Path(__file__).parent / "sample_autocorrect_output.csv"

DEEPXROMM_TEST_CODEC = os.environ.get("DEEPXROMM_TEST_CODEC", "avc1")


class TestRGBTrialProcess(unittest.TestCase):
    """Test function performance on an actual trial - RGB trial workflow"""

    def setUp(self):
        """Create trial with test data"""
        self.working_dir = Path.cwd() / "tmp"
        DeepXROMM.create_new_project(
            self.working_dir, mode="rgb", codec=DEEPXROMM_TEST_CODEC
        )

        # Make a trial directory
        trial_dir = self.working_dir / "trainingdata/test"
        trial_dir.mkdir(parents=True, exist_ok=True)

        # Make vars for pathing to find files easily
        self.trial_csv = trial_dir / "test.csv"
        self.rgb_path = trial_dir / "test_cam1.avi"
        self.cam2_path = trial_dir / "test_cam2.avi"

        # Move sample frame input to trainingdata
        shutil.copy("trial_slice.csv", str(self.trial_csv))
        shutil.copy("trial_cam1_slice.avi", str(self.rgb_path))
        shutil.copy("trial_cam2_slice.avi", str(self.cam2_path))
        self.deepxromm = DeepXROMM.load_project(self.working_dir)
        self.deepxromm.xma_to_dlc()

    def test_first_frame_matches_in_dlc_csv(self):
        """When I run xma_to_dlc, does the DLC CSV have the same data as my original file?"""

        xmalab_data = pd.read_csv(self.trial_csv)
        xmalab_first_row = xmalab_data.loc[0, :]

        # Load DLC data
        dlc_config = Path(self.deepxromm.config["path_config_file"])
        labeled_data_path = dlc_config.parent / "labeled-data/MyData"
        dlc_data = pd.read_hdf(labeled_data_path / "CollectedData_NA.h5")

        # Load DLC first row
        rgb_img_path = str(
            (labeled_data_path / "test_rgb_0001.png").relative_to(dlc_config.parent)
        )
        rgb_first_row = dlc_data.loc[rgb_img_path, :]

        for val in rgb_first_row.index:
            xmalab_key = f"{val[1]}_{val[2].upper()}"
            xmalab_data_point = xmalab_first_row[xmalab_key]
            dlc_data_point = rgb_first_row[val]
            with self.subTest(folder=xmalab_key):
                self.assertTrue(xmalab_data_point == dlc_data_point)

    def test_last_frame_matches_in_dlc_csv(self):
        """When I run xma_to_dlc, does the DLC CSV have the same data as my original file?"""
        self.deepxromm = DeepXROMM.load_project(self.working_dir)
        self.deepxromm.xma_to_dlc()

        # Load XMAlab data
        xmalab_data = pd.read_csv(self.trial_csv)

        # Load DLC data
        dlc_config = Path(self.deepxromm.config["path_config_file"])
        labeled_data_path = dlc_config.parent / "labeled-data/MyData"
        dlc_data = pd.read_hdf(labeled_data_path / "CollectedData_NA.h5")

        # Determine last frame included in training set
        last_file = Path(dlc_data.index[-1])
        last_frame_number = last_file.stem.split("_")[-1]
        last_frame_int = int(last_frame_number)

        # Load XMAlab last row
        xmalab_last_row = xmalab_data.loc[last_frame_int - 1]

        # Load DLC rgb last row
        rgb_img_path = str(
            (labeled_data_path / f"test_rgb_{last_frame_number}.png").relative_to(
                dlc_config.parent
            )
        )
        rgb_last_row = dlc_data.loc[rgb_img_path, :]

        for val in rgb_last_row.index:
            xmalab_key = f"{val[1]}_{val[2].upper()}"
            xmalab_data_point = xmalab_last_row[xmalab_key]
            dlc_data_point = rgb_last_row[val]
            with self.subTest(folder=xmalab_key):
                self.assertTrue(xmalab_data_point == dlc_data_point)

    def test_random_frame_matches_in_dlc_csv(self):
        """When I run xma_to_dlc, does the DLC CSV have the same data as my original file?"""
        self.deepxromm = DeepXROMM.load_project(self.working_dir)
        self.deepxromm.xma_to_dlc()

        # Load XMAlab data
        xmalab_data = pd.read_csv(self.trial_csv)

        # Load DLC data
        dlc_config = Path(self.deepxromm.config["path_config_file"])
        labeled_data_path = dlc_config.parent / "labeled-data/MyData"
        dlc_data = pd.read_hdf(labeled_data_path / "CollectedData_NA.h5")

        # Determine last frame included in training set
        file = Path(random.choice(dlc_data.index))
        frame_number = file.stem.split("_")[-1]
        frame_int = int(frame_number)

        # Load XMAlab last row
        xmalab_row = xmalab_data.loc[frame_int - 1]

        # Load DLC cam1 last row
        rgb_img_path = str(
            (labeled_data_path / f"test_rgb_{frame_number}.png").relative_to(
                dlc_config.parent
            )
        )
        rgb_row = dlc_data.loc[rgb_img_path, :]

        for val in rgb_row.index:
            xmalab_key = f"{val[1]}_{val[2].upper()}"
            xmalab_data_point = xmalab_row[xmalab_key]
            dlc_data_point = rgb_row[val]
            with self.subTest(folder=xmalab_key):
                self.assertTrue(xmalab_data_point == dlc_data_point)

    def test_create_training_dataset_succeeds(self):
        """Test that this project will create a dataset correctly"""
        self.deepxromm.create_training_dataset()

    def tearDown(self):
        """Remove the created temp project"""
        project_path = Path.cwd() / "tmp"
        shutil.rmtree(project_path)


def set_up_project(working_dir: Path):
    """RGB setup method to set up a project that's got xma_to_dlc and dlc_to_xma run on it"""
    DeepXROMM.create_new_project(working_dir, mode="rgb", codec=DEEPXROMM_TEST_CODEC)

    # Copy trial slice data
    trial_dir = working_dir / "trainingdata/test"
    trial_dir.mkdir(parents=True, exist_ok=True)
    trial_csv = trial_dir / "test.csv"
    rgb_path = trial_dir / "test_cam1.avi"
    cam2_path = trial_dir / "test_cam2.avi"

    shutil.copy("trial_slice.csv", str(trial_csv))
    shutil.copy("trial_cam1_slice.avi", str(rgb_path))
    shutil.copy("trial_cam2_slice.avi", str(cam2_path))

    # Run xma_to_dlc to create training dataset
    deepxromm = DeepXROMM.load_project(working_dir)
    deepxromm.xma_to_dlc()

    # Copy in mock DLC data
    rgb_df = pd.read_hdf("trial_rgbdlc.h5")
    output_dir = working_dir / "trials/test/it0"
    output_dir.mkdir(parents=True, exist_ok=True)
    mock_rgb_h5 = output_dir / "test_rgbDLC_resnet50_test_projectDec1shuffle1_100000.h5"
    mock_rgb_csv = (
        output_dir / "test_rgbDLC_resnet50_test_projectDec1shuffle1_100000.csv"
    )
    rgb_df.to_hdf(mock_rgb_h5, key="df_with_missing", mode="w")
    rgb_df.to_csv(mock_rgb_csv, na_rep="NaN")

    # Run DLC to XMA
    deepxromm.dlc_to_xma()
    return mock_rgb_h5, deepxromm


class TestDlcToXmaRGB(unittest.TestCase):
    """Test dlc_to_xma function in rgb mode with round-trip verification"""

    def setUp(self):
        """Create RGB project and generate mock DLC analysis output"""
        self.working_dir = Path.cwd() / "tmp"
        self.mock_rgb_h5, self.deepxromm = set_up_project(self.working_dir)

    def test_dlc_to_xma_creates_xmalab_format_files(self):
        """
        Given I have DLC analysis output with likelihood columns
        When I call dlc_to_xma
        Then it creates both CSV and HDF5 files
        """
        # Run dlc_to_xma
        output_dir = self.working_dir / "trials/test/it0"

        # Verify outputs exist
        xma_csv = output_dir / "test-Predicted2DPoints.csv"
        xma_h5 = output_dir / "test-Predicted2DPoints.h5"

        self.assertTrue(xma_csv.exists(), "XMAlab CSV not created")
        self.assertTrue(xma_h5.exists(), "XMAlab HDF5 not created")

    def test_dlc_to_xma_removes_likelihood_columns(self):
        """
        Given DLC analysis output has likelihood columns
        When I call dlc_to_xma
        Then the output XMAlab CSV has no likelihood columns
        And only contains X,Y coordinates
        """
        # Load and convert (split by camera)
        output_dir = self.working_dir / "trials/test/it0"

        # Check output
        xma_csv = output_dir / "test-Predicted2DPoints.csv"
        df = pd.read_csv(xma_csv)

        # Verify no likelihood columns
        for col in df.columns:
            self.assertNotIn("likelihood", col.lower())
            self.assertTrue(col.endswith(("_X", "_Y")))

    def test_first_frame_matches(self):
        """
        Given DLC-formatted output from a training run of this trial
        When I run dlc_to_xma
        Then the first frame of the reconstructed XMA data matches the DLC-formatted data
        """
        output_dir = self.working_dir / "trials/test/it0"
        rgb_dlc_data = pd.read_hdf(self.mock_rgb_h5)

        xmalab_data = pd.read_csv(output_dir / "test-Predicted2DPoints.csv")
        xmalab_first_row = xmalab_data.loc[0, :]

        dlc_config = Path(self.deepxromm.config["path_config_file"])
        labeled_data_path = dlc_config.parent / "labeled-data/MyData"
        rgb_first_row = rgb_dlc_data.loc[0, :]
        for val in rgb_first_row.index:
            if "likelihood" in val or "marker001" not in val or "x" not in val:
                continue
            xmalab_key = f"{val[1]}_{val[2].upper()}"
            xmalab_data_point = xmalab_first_row[xmalab_key]
            dlc_data_point = rgb_first_row[val]
            with self.subTest(folder=val):
                self.assertTrue(xmalab_data_point == dlc_data_point)

    def test_last_frame_matches(self):
        """
        Given DLC-formatted output from a training run of this trial
        When I run dlc_to_xma
        Then the last frame of the reconstructed XMA data matches the DLC-formatted data
        """
        output_dir = self.working_dir / "trials/test/it0"
        rgb_dlc_data = pd.read_hdf(
            output_dir / "test_rgbDLC_resnet50_test_projectDec1shuffle1_100000.h5"
        )

        xmalab_data = pd.read_csv(output_dir / "test-Predicted2DPoints.csv")
        # Find XMAlab last row
        xmalab_last_row_int = xmalab_data.index[-1]
        xmalab_last_row = xmalab_data.loc[xmalab_last_row_int, :]

        dlc_config = Path(self.deepxromm.config["path_config_file"])
        labeled_data_path = dlc_config.parent / "labeled-data/MyData"
        rgb_last_row = rgb_dlc_data.loc[xmalab_last_row_int, :]
        for val in rgb_last_row.index:
            if "likelihood" in val:
                continue
            xmalab_key = f"{val[1]}_{val[2].upper()}"
            xmalab_data_point = round(xmalab_last_row[xmalab_key], 4)
            dlc_data_point = round(rgb_last_row[val], 4)
            with self.subTest(folder=xmalab_key):
                self.assertTrue(xmalab_data_point == dlc_data_point)

    def test_random_frame_matches(self):
        """
        Given DLC-formatted output from a training run of this trial
        When I run dlc_to_xma
        Then a random frame of the reconstructed XMA data matches the DLC-formatted data
        """
        output_dir = self.working_dir / "trials/test/it0"
        rgb_dlc_data = pd.read_hdf(self.mock_rgb_h5)

        xmalab_data = pd.read_csv(output_dir / "test-Predicted2DPoints.csv")
        # Get random row from xmalab data
        xmalab_rand_row_int = random.choice(xmalab_data.index)
        xmalab_rand_row = xmalab_data.loc[xmalab_rand_row_int, :]

        dlc_config = Path(self.deepxromm.config["path_config_file"])
        labeled_data_path = dlc_config.parent / "labeled-data/MyData"
        rgb_rand_row = rgb_dlc_data.loc[xmalab_rand_row_int, :]
        for val in rgb_rand_row.index:
            if "likelihood" in val:
                continue
            xmalab_key = f"{val[1]}_{val[2].upper()}"
            xmalab_data_point = round(xmalab_rand_row[xmalab_key], 4)
            dlc_data_point = round(rgb_rand_row[val], 4)
            with self.subTest(folder=xmalab_key):
                self.assertTrue(xmalab_data_point == dlc_data_point)

    def tearDown(self):
        """Remove the created temp project"""
        project_path = Path.cwd() / "tmp"
        if project_path.exists():
            shutil.rmtree(project_path)


class TestAutocorrectRGB(unittest.TestCase):
    """Test that autocorrect runs properly for RGB project"""

    def setUp(self):
        """Create RGB project and generate mock DLC analysis output"""
        self.working_dir = Path.cwd() / "tmp"
        self.mock_rgb_h5, self.deepxromm = set_up_project(self.working_dir)

    def run_autocorrect(self):
        """Run autocorrect using the provided deepxromm project"""
        self.deepxromm.autocorrect_trials()

    def tearDown(self):
        """Remove the created temp project"""
        if self.working_dir.exists():
            shutil.rmtree(self.working_dir)



"""Converts data from XMALab into the format useful for DLC training."""

import logging
import tempfile
from pathlib import Path

from subprocess import Popen, PIPE

import blend_modes
import cv2
from PIL import Image
import numpy as np
import pandas as pd
import random
import yaml

from deepxromm.xrommtools import dlc_to_xma

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


class XMADataProcessor:
    """Converts data from XMALab into the format useful for DLC training."""

    def __init__(self, config):
        self._config = config
        self._swap_markers = config["swapped_markers"]
        self._cross_markers = config["crossed_markers"]

    def dlc_to_xma(self):
        """Convert DLC-formatted training output into XMAlab-formatted data"""
        mode = self._config["mode"]
        trials = self.list_trials()

        if mode in ["2D", "per_cam"]:
            correct_function_signature = dlc_to_xma
        elif mode == "rgb":
            correct_function_signature = self._split_dlc_to_xma
        else:
            raise AttributeError(f"Unsupported mode: {mode}")

        with open(self._config["path_config_file"], "r") as dlc_config:
            dlc_proj = yaml.safe_load(dlc_config)

        iteration = int(dlc_proj["iteration"])
        for trial in trials:
            correct_function_signature(trial, iteration)

    def _split_dlc_to_xma(self, trial_path: Path, iteration: int, save_hdf=True):
        """Takes the output from RGB deeplabcut and splits it into XMAlab-readable output"""
        current_files = trial_path.glob("*")
        logger.debug(f"Current files in directory {current_files}")
        # Use the CSV from the user's training data to fetch bodyparts
        sample_trial_path = self.list_trials("trainingdata")[0]
        trial_csv_path = self.find_trial_csv(sample_trial_path)
        rgb_parts = self.get_bodyparts_from_xma(trial_csv_path, mode="rgb")

        bodyparts_xy = []
        for part in rgb_parts:
            bodyparts_xy.append(part + "_X")
            bodyparts_xy.append(part + "_Y")

        iteration_folder = trial_path / f"it{iteration}"
        csv_path = self.find_trial_csv(
            iteration_folder, "rgbDLC"
        )  # Assumes that the project itself doesn't have rgbDLC in this format in it
        trial = trial_path.name
        xma_csv_path = iteration_folder / f"{trial}-Predicted2DPoints.csv"

        df = pd.read_csv(csv_path, skiprows=1)
        df.index = df["bodyparts"]
        df = df.drop(columns=df.columns[df.loc["coords"] == "likelihood"])
        df = df.drop(
            columns=[
                column
                for column in df.columns
                if column not in rgb_parts
                and column not in [f"{bodypart}.1" for bodypart in rgb_parts]
            ]
        )
        df.columns = bodyparts_xy
        df = df.drop(index="coords")
        df.to_csv(xma_csv_path, index=False)
        print(
            "Successfully split DLC format to XMALab 2D points; saved "
            + str(xma_csv_path)
        )
        if save_hdf:
            tracked_hdf = xma_csv_path.with_suffix(".h5")
            logger.debug(f"Tracked hdf stored at {str(tracked_hdf)}")
            df.to_hdf(
                tracked_hdf, "df_with_missing", format="table", mode="w", nan_rep="NaN"
            )

    def validate_codec(self, codec: str, width: int = 640, height: int = 480) -> bool:
        """
        Validate if a video codec is available on the current system.

        Args:
            codec: FourCC codec code (e.g., "avc1", "DIVX", "XVID")
            width: Test video width (default 640)
            height: Test video height (default 480)

        Returns:
            True if codec is available and functional, False otherwise

        Note:
            Special cases "uncompressed" and 0 always return True as they
            use different encoding mechanisms.
        """
        # Special cases that don't use cv2.VideoWriter with fourcc
        if codec == "uncompressed" or codec == 0:
            return True

        # Test codec by creating a temporary VideoWriter
        with tempfile.TemporaryDirectory() as tmpdir:
            try:
                test_path = Path(tmpdir) / "codec_test.avi"
                fourcc = cv2.VideoWriter_fourcc(*codec)
                writer = cv2.VideoWriter(
                    str(test_path),
                    fourcc,
                    1.0,  # Low FPS for test
                    (width, height),
                )

                # Check if writer opened successfully
                is_valid = writer.isOpened()

                # Try writing a test frame to ensure codec actually works
                if is_valid:
                    test_frame = np.zeros((height, width, 3), dtype=np.uint8)
                    writer.write(test_frame)

                writer.release()
                logger.info(
                    f"Codec validation for '{codec}': {'PASSED' if is_valid else 'FAILED'}"
                )
                return is_valid

            except Exception as e:
                logger.warning(
                    f"Codec validation for '{codec}' failed with exception: {e}"
                )
                return False

    def get_codec_error_message(self, failed_codec: str, operation: str) -> str:
        """
        Generate descriptive error message for codec validation failure.

        Args:
            failed_codec: The codec that failed validation
            operation: Operation type ("split" or "merge")

        Returns:
            Formatted error message with suggestions
        """
        return f"""
Video codec '{failed_codec}' is not available on this system for {operation}_rgb operation.

Common alternative codecs to try:
  - "avc1"  : H.264 codec (best quality, not always available)
  - "DIVX"  : DivX codec (widely available)
  - "XVID"  : Xvid codec (widely available)
  - "mp4v"  : MPEG-4 codec (generally available)
  - "MJPG"  : Motion JPEG (always available, larger file sizes)
  - "uncompressed" : Raw video via ffmpeg (largest files, highest quality)

To change the codec, update your project config file:
  video_codec: "DIVX"  # Change this line to one of the alternatives above

Note: Codec availability depends on your OpenCV build and system codecs.
""".strip()

    def find_trial_csv(self, trial_path: Path, identifier: str | None = None) -> Path:
        """
        Takes the path to a trial and returns the path to a trial CSV.
        Errors if there is not exactly 1 trial CSV in a trial folder.
        """
        if identifier is not None:
            csv_path = list(trial_path.glob(f"*{identifier}*.csv"))
        else:
            csv_path = list(trial_path.glob("*.csv"))

        if len(csv_path) > 1:
            print(csv_path)
            raise FileExistsError(
                f"Found more than 1 CSV file with identifier {identifier} for trial: {trial_path}"
            )
        if len(csv_path) <= 0:
            logger.debug(f"Current files in {str(trial_path)}")
            logger.debug(list(trial_path.glob("*")))
            raise FileNotFoundError(
                f"Couldn't find a CSV file with identifier {identifier} for trial: {trial_path}"
            )

        return csv_path[0]

    def get_bodyparts_from_xma(self, csv_path: Path, mode: str):
        """Takes the filepath of an XMAlab CSV file and returns marker names"""

        trial_csv = pd.read_csv(
            csv_path,
            sep=",",
            header=0,
            dtype="float",
            na_values="NaN",
        )
        names = trial_csv.columns.values
        if mode == "rgb":
            parts = [name.rsplit("_", 1)[0] for name in names]
            if self._swap_markers:
                parts = parts + [f"sw_{part}" for part in parts]
            if self._cross_markers:
                parts = parts + [
                    f"cx_{part}_cam1x2"
                    for part in [name.rsplit("_", 2)[0] for name in names]
                ]
        elif mode in ["2D", "per_cam"]:
            parts = [name.rsplit("_", 2)[0] for name in names]
        else:
            raise SyntaxError("Invalid value for mode parameter")

        # I do it this way to maintain ordering in the list, since that's
        # important for DeepLabCut
        parts_unique = []
        for part in parts:
            if part not in parts_unique:
                parts_unique.append(part)
        return parts_unique

    def make_rgb_videos(self, suffix: str):
        """For all trials in given data path merges 2 videos into single RGB video."""
        trials = self.list_trials(suffix=suffix)
        for path_to_trial in trials:
            self._merge_rgb(path_to_trial)

    def xma_to_dlc_rgb(self, suffix: str):
        """Convert XMAlab input into RGB-ready DLC input"""
        trials = self.list_trials(suffix=suffix)
        for trial_path in trials:
            df = pd.read_csv(self.find_trial_csv(trial_path))
            substitute_data_relpath = (
                Path("labeled-data") / self._config["dataset_name"]
            )
            dlc_config_path = Path(self._config["path_config_file"])
            substitute_data_abspath = dlc_config_path / substitute_data_relpath
            df = df.dropna(how="all")
            list_of_frames = df.index + 1
            self._extract_matched_frames_rgb(
                trial_path,
                substitute_data_abspath,
                list_of_frames,
            )
            self._splice_xma_to_dlc(trial_path)

    def find_cam_file(self, path: Path, identifier: str):
        """Searches a file for a given cam video in the trail folder."""
        files = list(path.glob(f"*{identifier}*.avi"))
        logger.debug(files)
        if files:
            result = files[0]
            print(f"Found file {result} for {identifier}")
            return files[0]
        else:
            raise FileNotFoundError(
                f"No video file found containing '{identifier}' in {path}"
            )

    def list_trials(self, suffix: str = "trials") -> list[Path]:
        """
        Get list of trial directories from project working directory.

        This is the ONLY method for listing trials. All trial access is validated
        against the project's working_dir to prevent unauthorized file access.

        Args:
            suffix: Relative path from working_dir to trials folder.
                    Examples: "trials", "trainingdata", "data/experiments"
                    Cannot contain '..' or start with '/' for security.

        Returns:
            List of trial directory Path objects (excludes hidden folders)

        Raises:
            FileNotFoundError: If no trials found in working_dir/suffix
            ValueError: If suffix contains path traversal attempts or absolute paths

        Security:
            Path traversal attempts (../, absolute paths) are rejected to ensure
            all file access remains within the project working_dir.
        """
        # Security validation: prevent directory traversal
        if ".." in suffix:
            raise ValueError(
                f"Security error: Path traversal detected in suffix '{suffix}'. "
                "Suffix cannot contain '..' for security reasons."
            )
        if suffix.startswith("/"):
            raise ValueError(
                f"Security error: Absolute path detected in suffix '{suffix}'. "
                "Suffix must be a relative path within working_dir."
            )

        working_dir = Path(self._config["working_dir"])
        trial_path = working_dir / suffix

        # Additional security check: ensure resolved path is within working_dir
        try:
            resolved_trial_path = trial_path.resolve()
            resolved_working_dir = working_dir.resolve()
            if not str(resolved_trial_path).startswith(str(resolved_working_dir)):
                raise ValueError(
                    f"Security error: Path traversal detected. "
                    f"Resolved path '{resolved_trial_path}' is outside working_dir."
                )
        except (OSError, RuntimeError) as e:
            raise ValueError(f"Invalid path in suffix '{suffix}': {e}")

        # Get list of trial directories (excluding hidden folders)
        trialnames = [
            folder
            for folder in trial_path.glob("*")
            if (trial_path / folder).is_dir() and not folder.name.startswith(".")
        ]

        if len(trialnames) == 0:
            raise FileNotFoundError(
                f"No trials found in {trial_path}. "
                "Please ensure trial directories exist and are not hidden."
            )

        return trialnames

    def split_rgb(self, trial_path: Path, codec=None):
        """Takes a RGB video with different grayscale data written to the R, G, and B channels and splits it back into its component source videos."""
        # Use provided codec, otherwise fall back to config value
        if codec is None:
            codec = self._config.get("video_codec", "avc1")

        trial_name = trial_path.name
        out_name = trial_name + "_split_"

        rgb_video_path = trial_path / f"{trial_name}_rgb.avi"
        try:
            rgb_video = cv2.VideoCapture(rgb_video_path)
        except FileNotFoundError as e:
            print(f"Couldn't find video at {rgb_video_path}")
            raise e
        frame_width = int(rgb_video.get(3))
        frame_height = int(rgb_video.get(4))
        frame_rate = round(rgb_video.get(5), 2)

        # Validate codec is available on this system (unless using uncompressed)
        if not self.validate_codec(codec, frame_width, frame_height):
            error_msg = self.get_codec_error_message(codec, "split")
            raise RuntimeError(error_msg)

        if codec == "uncompressed":
            pix_format = "gray"  ##change to 'yuv420p' for color or 'gray' for grayscale. 'pal8' doesn't play on macs
            cam1_split_ffmpeg = Popen(
                [
                    "ffmpeg",
                    "-y",
                    "-f",
                    "image2pipe",
                    "-vcodec",
                    "png",
                    "-r",
                    str(int(frame_rate)),
                    "-i",
                    "-",
                    "-vcodec",
                    "rawvideo",
                    "-pix_fmt",
                    pix_format,
                    "-r",
                    str(int(frame_rate)),
                    f"{trial_path}/{out_name}" + "cam1.avi",
                ],
                stdin=PIPE,
            )
            cam2_split_ffmpeg = Popen(
                [
                    "ffmpeg",
                    "-y",
                    "-f",
                    "image2pipe",
                    "-vcodec",
                    "png",
                    "-r",
                    str(int(frame_rate)),
                    "-i",
                    "-",
                    "-vcodec",
                    "rawvideo",
                    "-pix_fmt",
                    pix_format,
                    "-r",
                    str(int(frame_rate)),
                    f"{trial_path}/{out_name}" + "cam2.avi",
                ],
                stdin=PIPE,
            )
            blue_split_ffmpeg = Popen(
                [
                    "ffmpeg",
                    "-y",
                    "-f",
                    "image2pipe",
                    "-vcodec",
                    "png",
                    "-r",
                    str(int(frame_rate)),
                    "-i",
                    "-",
                    "-vcodec",
                    "rawvideo",
                    "-pix_fmt",
                    pix_format,
                    "-r",
                    str(int(frame_rate)),
                    f"{trial_path}/{out_name}" + "blue.avi",
                ],
                stdin=PIPE,
            )
        else:
            if codec == 0:
                fourcc = 0
            else:
                fourcc = cv2.VideoWriter_fourcc(*codec)
            cam1 = cv2.VideoWriter(
                f"{trial_path}/{out_name}" + "cam1.avi",
                fourcc,
                frame_rate,
                (frame_width, frame_height),
            )
            cam2 = cv2.VideoWriter(
                f"{trial_path}/{out_name}" + "cam2.avi",
                fourcc,
                frame_rate,
                (frame_width, frame_height),
            )
            blue_channel = cv2.VideoWriter(
                f"{trial_path}/{out_name}" + "blue.avi",
                fourcc,
                frame_rate,
                (frame_width, frame_height),
            )

            # Verify VideoWriters opened successfully
            if not cam1.isOpened():
                raise RuntimeError(
                    f"Failed to create cam1 video writer with codec '{codec}'"
                )
            if not cam2.isOpened():
                raise RuntimeError(
                    f"Failed to create cam2 video writer with codec '{codec}'"
                )
            if not blue_channel.isOpened():
                raise RuntimeError(
                    f"Failed to create blue channel video writer with codec '{codec}'"
                )

        i = 1
        while rgb_video.isOpened():
            ret, frame = rgb_video.read()
            print(f"Current Frame: {i}")
            i = i + 1
            if ret:
                B, G, R = cv2.split(frame)
                if codec == "uncompressed":
                    im_r = Image.fromarray(R)
                    im_g = Image.fromarray(G)
                    im_b = Image.fromarray(B)
                    # TODO - add error handling/assertions
                    im_r.save(cam1_split_ffmpeg.stdin, "PNG")
                    im_g.save(cam2_split_ffmpeg.stdin, "PNG")
                    im_b.save(blue_split_ffmpeg.stdin, "PNG")
                    if cv2.waitKey(1) & 0xFF == ord("q"):
                        break
                else:
                    cam1.write(R)
                    cam2.write(G)
                    blue_channel.write(B)
                    if cv2.waitKey(1) & 0xFF == ord("q"):
                        break
            else:
                break
        if codec == "uncompressed":
            cam1_split_ffmpeg.stdin.close()
            cam1_split_ffmpeg.wait()
            cam2_split_ffmpeg.stdin.close()
            cam2_split_ffmpeg.wait()
            blue_split_ffmpeg.stdin.close()
            blue_split_ffmpeg.wait()
        else:
            cam1.release()
            cam2.release()
            blue_channel.release()
        rgb_video.release()
        cv2.destroyAllWindows()
        print(f"Cam1 grayscale video created at {trial_path}/{out_name}cam1.avi!")
        print(f"Cam2 grayscale video created at {trial_path}/{out_name}cam2.avi!")
        print(
            f"Blue channel grayscale video created at {trial_path}/{out_name}blue.avi!"
        )

    def _merge_rgb(self, trial_path: Path, codec=None, mode="difference"):
        """
        Takes the path to a trial subfolder and exports a single new video with
        cam1 video written to the red channel and cam2 video written to the
        green channel. The blue channel is, depending on the value of config
        "mode", either the difference blend between A and B, the multiply
        blend, or just a black frame.
        """
        # Use provided codec, otherwise fall back to config value
        if codec is None:
            codec = self._config.get("video_codec", "avc1")

        print("Merging RGBs")
        trial_name = trial_path.name
        rgb_video_path = trial_path / f"{trial_name}_rgb.avi"
        if rgb_video_path.exists():
            print("RGB video already created. Skipping.")
            return
        cam1_video_path = self.find_cam_file(trial_path, "cam1")
        cam1_video = cv2.VideoCapture(cam1_video_path)

        cam2_video_path = self.find_cam_file(trial_path, "cam2")
        cam2_video = cv2.VideoCapture(cam2_video_path)

        frame_width = int(cam1_video.get(3))
        frame_height = int(cam1_video.get(4))
        frame_rate = round(cam1_video.get(5), 2)

        # Validate codec is available on this system
        if not self.validate_codec(codec, frame_width, frame_height):
            error_msg = self.get_codec_error_message(codec, "merge")
            raise RuntimeError(error_msg)

        # Note: "uncompressed" codec is not supported for merge_rgb
        # If needed in the future, implement ffmpeg pipeline like in split_rgb
        if codec == "uncompressed":
            raise RuntimeError(
                "The 'uncompressed' codec is not currently supported for merge_rgb operation. "
                "Please use a compressed codec like 'avc1', 'DIVX', 'XVID', 'mp4v', or 'MJPG'."
            )

        if codec == 0:
            fourcc = 0
        else:
            fourcc = cv2.VideoWriter_fourcc(*codec)
        out = cv2.VideoWriter(
            f"{trial_path}/{trial_name}_rgb.avi",
            fourcc,
            frame_rate,
            (frame_width, frame_height),
        )

        # Verify VideoWriter opened successfully
        if not out.isOpened():
            raise RuntimeError(
                f"Failed to create RGB video writer with codec '{codec}'"
            )

        i = 1
        while cam1_video.isOpened():
            print(f"Current Frame: {i}")
            ret_cam1, frame_cam1 = cam1_video.read()
            _, frame_cam2 = cam2_video.read()
            if ret_cam1:
                frame_cam1 = cv2.cvtColor(frame_cam1, cv2.COLOR_BGR2BGRA, 4).astype(
                    np.float32
                )
                frame_cam2 = cv2.cvtColor(frame_cam2, cv2.COLOR_BGR2BGRA, 4).astype(
                    np.float32
                )
                frame_cam1 = cv2.normalize(
                    frame_cam1, None, 0, 255, norm_type=cv2.NORM_MINMAX
                )
                frame_cam2 = cv2.normalize(
                    frame_cam2, None, 0, 255, norm_type=cv2.NORM_MINMAX
                )
                if mode == "difference":
                    extra_channel = blend_modes.difference(frame_cam1, frame_cam2, 1)
                elif mode == "multiply":
                    extra_channel = blend_modes.multiply(frame_cam1, frame_cam2, 1)
                else:
                    extra_channel = np.zeros((frame_width, frame_height, 3), np.uint8)
                    extra_channel = cv2.cvtColor(
                        extra_channel, cv2.COLOR_BGR2BGRA, 4
                    ).astype(np.float32)
                frame_cam1 = cv2.cvtColor(frame_cam1, cv2.COLOR_BGRA2BGR).astype(
                    np.uint8
                )
                frame_cam2 = cv2.cvtColor(frame_cam2, cv2.COLOR_BGRA2BGR).astype(
                    np.uint8
                )
                extra_channel = cv2.cvtColor(extra_channel, cv2.COLOR_BGRA2BGR).astype(
                    np.uint8
                )
                frame_cam1 = cv2.cvtColor(frame_cam1, cv2.COLOR_BGR2GRAY)
                frame_cam2 = cv2.cvtColor(frame_cam2, cv2.COLOR_BGR2GRAY)
                extra_channel = cv2.cvtColor(extra_channel, cv2.COLOR_BGR2GRAY)
                merged = cv2.merge((extra_channel, frame_cam2, frame_cam1))
                out.write(merged)
                if cv2.waitKey(1) & 0xFF == ord("q"):
                    break
            else:
                break

            i = i + 1
        cam1_video.release()
        cam2_video.release()
        out.release()
        cv2.destroyAllWindows()
        print(f"Merged RGB video created at {trial_path}/{trial_name}_rgb.avi!")

    def _splice_xma_to_dlc(self, trial_path: Path, outlier_mode=False):
        """Takes csv of XMALab 2D XY coordinates from 2 cameras, outputs spliced hdf+csv data for DeepLabCut"""
        dlc_path = Path(self._config["path_config_file"]).parent
        trial_name = trial_path.name
        substitute_data_abspath = (
            dlc_path / "labeled-data" / self._config["dataset_name"]
        )
        trial_csv_path = self.find_trial_csv(trial_path)
        markers = self.get_bodyparts_from_xma(trial_csv_path, mode="2D")

        # TODO: this entire section can be solved with a creative call to
        # get_bodyparts_from_xma and some dataFrame manipulation to be
        # significantly shorter (and potentially faster?)
        try:
            df = pd.read_csv(trial_csv_path)
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Cannot find your trainingdata 2DPoints csv file is named .csv"
            ) from e
        if self._swap_markers:
            print("Creating cam1Y-cam2Y-swapped synthetic markers")
            swaps = []
            df_sw = pd.DataFrame()
            for marker in markers:
                name_x1 = marker + "_cam1_X"
                name_x2 = marker + "_cam2_X"
                name_y1 = marker + "_cam1_Y"
                name_y2 = marker + "_cam2_Y"
                swap_name_x1 = "sw_" + name_x1
                swap_name_x2 = "sw_" + name_x2
                swap_name_y1 = "sw_" + name_y1
                swap_name_y2 = "sw_" + name_y2
                df_sw[swap_name_x1] = df[name_x1]
                df_sw[swap_name_y1] = df[name_y2]
                df_sw[swap_name_x2] = df[name_x2]
                df_sw[swap_name_y2] = df[name_y1]
                swaps.extend([swap_name_x1, swap_name_y1, swap_name_x2, swap_name_y2])
            df = df.join(df_sw)
            print(swaps)
        if self._cross_markers:
            print("Creating cam1-cam2-crossed synthetic markers")
            crosses = []
            df_cx = pd.DataFrame()
            for marker in markers:
                name_x1 = marker + "_cam1_X"
                name_x2 = marker + "_cam2_X"
                name_y1 = marker + "_cam1_Y"
                name_y2 = marker + "_cam2_Y"
                cross_name_x = "cx_" + marker + "_cam1x2_X"
                cross_name_y = "cx_" + marker + "_cam1x2_Y"
                df_cx[cross_name_x] = df[name_x1] * df[name_x2]
                df_cx[cross_name_y] = df[name_y1] * df[name_y2]
                crosses.extend([cross_name_x, cross_name_y])
            df = df.join(df_cx)
            print(crosses)
        names_final = df.columns.values
        parts_final = [name.rsplit("_", 1)[0] for name in names_final]
        parts_unique_final = []
        for part in parts_final:
            if not part in parts_unique_final:
                parts_unique_final.append(part)
        print("Importing markers: ")
        print(parts_unique_final)
        with open(self._config["path_config_file"], "r") as dlc_config:
            dlc_proj = yaml.safe_load(dlc_config)

        dlc_proj["bodyparts"] = parts_unique_final

        with open(self._config["path_config_file"], "w") as dlc_config:
            yaml.dump(dlc_proj, dlc_config, sort_keys=False)

        df = df.dropna(how="all")
        list_of_frames = df.index + 1
        unique_frames_set = set(list_of_frames)

        # Ensure that all frames were unique originally
        assert len(list_of_frames) == len(unique_frames_set)

        unique_frames = sorted(unique_frames_set)
        print("Importing frames: ")
        print(unique_frames)
        df["frame_index"] = [
            str(
                (
                    substitute_data_abspath
                    / f"{trial_name}_rgb_{str(index).zfill(4)}.png"
                ).relative_to(dlc_path)
            )
            for index in unique_frames
        ]
        df["scorer"] = self._config["experimenter"]
        df = df.melt(id_vars=["frame_index", "scorer"])
        new = df["variable"].str.rsplit("_", n=1, expand=True)
        df["variable"], df["coords"] = new[0], new[1]
        df = df.rename(columns={"variable": "bodyparts"})
        df["coords"] = df["coords"].str.rstrip(" ").str.lower()
        cat_type = pd.api.types.CategoricalDtype(
            categories=parts_unique_final, ordered=True
        )
        df["bodyparts"] = df["bodyparts"].str.lstrip(" ").astype(cat_type)
        newdf = df.pivot_table(
            columns=["scorer", "bodyparts", "coords"],
            index="frame_index",
            values="value",
            aggfunc="first",
            dropna=False,
        )
        newdf.index.name = None
        if not substitute_data_abspath.exists():
            substitute_data_abspath.mkdir(parents=True, exist_ok=True)
        if outlier_mode:
            tracked_hdf = substitute_data_abspath / "MachineLabelsRefine.h5"
        else:
            data_name = "CollectedData_" + self._config["experimenter"] + ".h5"
            tracked_hdf = substitute_data_abspath / data_name

        print(tracked_hdf)
        newdf.to_hdf(tracked_hdf, "df_with_missing", format="table", mode="w")
        tracked_csv = tracked_hdf.with_suffix(".csv")
        newdf.to_csv(tracked_csv, na_rep="NaN")
        print(
            "Successfully spliced XMALab 2D points to DLC format",
            f"saved {tracked_hdf}",
            f"saved {tracked_csv}",
            sep="\n",
        )

    def _extract_matched_frames_rgb(
        self, trial_path, labeled_data_path, indices, compression=1
    ):
        """Given a list of frame indices and a project path, produce a folder (in labeled-data) of matching frame pngs per source video.
        Optionally, compress the output PNGs. Factor ranges from 0 (no compression) to 9 (most compression)
        """
        extracted_frames = []
        trainingdata_path = Path(self._config["working_dir"]) / "trainingdata"
        trial_name = trial_path.name
        video_path = trainingdata_path / trial_name / f"{trial_name}_rgb.avi"
        dlc_path = Path(self._config["path_config_file"]).parent
        labeled_data_path = dlc_path / "labeled-data" / self._config["dataset_name"]
        # Convert 1-indexed frame numbers to 0-indexed for new method
        zero_indexed_frames = [idx - 1 for idx in indices]
        frames_from_vid = self.extract_frames_from_video(
            source_path=video_path,
            frame_indices=zero_indexed_frames,
            output_dir=labeled_data_path,
            output_name_base=video_path.parent.name,  # Gets trial name from filename
            mode="rgb",
            compression=compression,
        )
        extracted_frames.append(frames_from_vid)
        print("Extracted " + str(len(indices)) + f" matching frames from {video_path}")

    def _extract_frame_selection_loop(self, idx: list, nframes: int):
        """Extract the existing frame selection algorithm (preserve existing algorithm)

        Args:
            idx: List of frame indices for each trial
            nframes: Number of frames to select

        Returns:
            List of selected frame indices for each trial
        """
        picked_frames = []

        # Check if we have enough frames
        if sum(len(x) for x in idx) < nframes:
            raise ValueError("nframes is bigger than number of detected frames")

        # Pick frames to extract (NOTE this is random currently)
        # current code iteratively picks one frame at a time from each shuffled trial until # of picked_frames hits nframes
        count = 0
        while sum(len(x) for x in picked_frames) < nframes:
            for trialnum in range(len(idx)):
                if sum(len(x) for x in picked_frames) < nframes:
                    if count == 0:
                        picked_frames.insert(trialnum, [idx[trialnum][count]])
                    elif count < len(idx[trialnum]):
                        picked_frames[trialnum] = picked_frames[trialnum] + [
                            idx[trialnum][count]
                        ]
                count += 1

        return picked_frames

    def extract_frames_from_video(
        self,
        source_path: Path,
        frame_indices: list[int],
        output_dir: Path,
        output_name_base: str,
        mode: str = "2D",
        camera: int | None = None,
        compression: int = 0,
    ) -> list[str]:
        """
        Extract specified frames from video or image folder.

        Unified interface for all frame extraction needs across different modes.

        Args:
            source_path: Path to video file or image directory
            frame_indices: List of frame numbers to extract (0-indexed)
            output_dir: Directory to save extracted frames
            output_name_base: Base name for output files (e.g., trial name)
            mode: Extraction mode - determines filename format
                  "rgb": {output_name_base}_rgb_{frame}.png
                  "2D": {output_name_base}_cam{camera}_{frame}.png
                  "per_cam": {output_name_base}_{frame}.png
            camera: Camera number (required for 2D mode)
            compression: PNG compression level (0=none, 9=max)

        Returns:
            List of absolute file paths to extracted frames (as strings)

        Raises:
            ValueError: If mode is invalid or required parameters are missing
            FileNotFoundError: If source_path doesn't exist
        """
        # Validate mode
        valid_modes = ["2D", "per_cam", "rgb"]
        if mode not in valid_modes:
            raise ValueError(f"Invalid mode '{mode}'. Must be one of: {valid_modes}")

        # Validate camera parameter for 2D mode
        if mode == "2D" and camera is None:
            raise ValueError("Camera parameter is required for 2D mode")

        # Validate compression level
        if not (0 <= compression <= 9):
            raise ValueError(f"Compression must be between 0 and 9, got {compression}")

        # Validate source exists
        if not source_path.exists():
            raise FileNotFoundError(f"Source path not found: {source_path}")

        # Create output directory if needed
        output_dir.mkdir(parents=True, exist_ok=True)

        # Route to appropriate handler based on source type
        if source_path.is_dir():
            return self._extract_from_image_folder(
                source_path,
                frame_indices,
                output_dir,
                output_name_base,
                mode,
                camera,
            )
        else:
            return self._extract_from_video_file(
                source_path,
                frame_indices,
                output_dir,
                output_name_base,
                mode,
                camera,
                compression,
            )

    def _extract_from_video_file(
        self,
        video_path: Path,
        frame_indices: list[int],
        output_dir: Path,
        output_name_base: str,
        mode: str,
        camera: int | None,
        compression: int,
    ) -> list[str]:
        """
        Extract frames from video file.

        Args:
            video_path: Path to video file
            frame_indices: List of 0-indexed frame numbers to extract
            output_dir: Directory to save frames
            output_name_base: Base name for output files
            mode: Extraction mode ("2D", "per_cam", or "rgb")
            camera: Camera number (for 2D mode)
            compression: PNG compression level (0-9)

        Returns:
            List of absolute file paths as strings
        """
        frame_indices_set = set(frame_indices)
        last_frame_to_analyze = max(frame_indices)
        extracted_paths = []

        cap = cv2.VideoCapture(str(video_path))
        if not cap.isOpened():
            raise IOError(f"Failed to open video file: {video_path}")

        frame_index = 0
        while cap.isOpened():
            ret, frame = cap.read()
            if not ret:
                break

            if frame_index > last_frame_to_analyze:
                break

            if frame_index in frame_indices_set:
                # Print progress with human-readable (1-indexed) frame number
                print(f"Extracting frame {frame_index + 1}")

                # Build filename based on mode (use 1-based indexing for file names)
                frame_str = str(frame_index + 1).zfill(4)
                if mode == "rgb":
                    filename = f"{output_name_base}_rgb_{frame_str}.png"
                elif mode == "2D":
                    filename = f"{output_name_base}_cam{camera}_{frame_str}.png"
                else:  # per_cam
                    filename = f"{output_name_base}_{frame_str}.png"

                output_path = output_dir / filename
                cv2.imwrite(
                    str(output_path), frame, [cv2.IMWRITE_PNG_COMPRESSION, compression]
                )
                extracted_paths.append(str(output_path.absolute()))

            frame_index += 1

        cap.release()
        cv2.destroyAllWindows()

        return extracted_paths

    def _extract_from_image_folder(
        self,
        img_folder: Path,
        frame_indices: list[int],
        output_dir: Path,
        output_name_base: str,
        mode: str,
        camera: int | None,
    ) -> list[str]:
        """
        Extract frames from image folder.

        Args:
            img_folder: Path to folder containing images
            frame_indices: List of 0-indexed frame numbers to extract
            output_dir: Directory to save frames
            output_name_base: Base name for output files
            mode: Extraction mode ("2D", "per_cam", or "rgb")
            camera: Camera number (for 2D mode)

        Returns:
            List of absolute file paths as strings
        """
        extracted_paths = []
        imgs = sorted(list(img_folder.glob("*")))
        frame_indices_sorted = sorted(frame_indices)

        for frame_idx in frame_indices_sorted:
            if frame_idx >= len(imgs):
                logger.warning(
                    f"Frame index {frame_idx} out of range (folder has {len(imgs)} images)"
                )
                continue

            # Print progress with human-readable (1-indexed) frame number
            print(f"Extracting frame {frame_idx + 1}")

            # Load image
            img_path = imgs[frame_idx]
            image = cv2.imread(str(img_path))

            if image is None:
                logger.warning(f"Failed to load image: {img_path}")
                continue

            # Build filename based on mode
            # Note: For image folders, we use frame_idx + 1 for human readability
            frame_str = str(frame_idx + 1).zfill(4)
            if mode == "rgb":
                filename = f"{output_name_base}_rgb_{frame_str}.png"
            elif mode == "2D":
                filename = f"{output_name_base}_cam{camera}_{frame_str}.png"
            else:  # per_cam
                filename = f"{output_name_base}_{frame_str}.png"

            output_path = output_dir / filename
            cv2.imwrite(str(output_path), image)
            extracted_paths.append(str(output_path.absolute()))

        return extracted_paths

    def build_dlc_dataframe(
        self,
        data: pd.DataFrame,
        scorer: str,
        relnames: list[str],
        pointnames: list[str],
    ) -> pd.DataFrame:
        """
        Build DeepLabCut MultiIndex DataFrame from raw coordinate data.

        Args:
            data: DataFrame with coordinates (columns are alternating x,y for each point)
            scorer: Name of the scorer/experimenter
            relnames: List of absolute image file paths (for index)
            pointnames: List of body part names

        Returns:
            DataFrame with MultiIndex columns (scorer, bodyparts, coords)

        Raises:
            ValueError: If data dimensions don't match expected structure
        """
        # Validate inputs
        expected_cols = len(pointnames) * 2  # x,y for each point
        if data.shape[1] != expected_cols:
            raise ValueError(
                f"Data has {data.shape[1]} columns but expected {expected_cols} "
                f"for {len(pointnames)} body parts (2 coords each)"
            )

        if data.shape[0] != len(relnames):
            raise ValueError(
                f"Data has {data.shape[0]} rows but {len(relnames)} image names provided"
            )

        # Build MultiIndex DataFrame
        dataFrame = pd.DataFrame()
        temp = np.empty((data.shape[0], 2))
        temp[:] = np.nan

        for idx, relname in enumerate(relnames):
            if "_cam2" in relname and self._config["mode"] == "per_cam":
                relnames[idx] = str(
                    Path(relname).relative_to(
                        Path(self._config["path_config_file_2"]).parent
                    )
                )
            else:
                relnames[idx] = str(
                    Path(relname).relative_to(
                        Path(self._config["path_config_file"]).parent
                    )
                )

        for i, bodypart in enumerate(pointnames):
            index = pd.MultiIndex.from_product(
                [[scorer], [bodypart], ["x", "y"]],
                names=["scorer", "bodyparts", "coords"],
            )
            frame = pd.DataFrame(temp, columns=index, index=relnames)
            frame.iloc[:, 0:2] = data.iloc[:, 2 * i : 2 * i + 2].values.astype(float)
            dataFrame = pd.concat([dataFrame, frame], axis=1)

        # Clean various NaN representations
        dataFrame.replace("", np.nan, inplace=True)
        dataFrame.replace(" NaN", np.nan, inplace=True)
        dataFrame.replace(" NaN ", np.nan, inplace=True)
        dataFrame.replace("NaN ", np.nan, inplace=True)
        dataFrame = dataFrame.apply(pd.to_numeric, errors="coerce")

        return dataFrame

    def save_dlc_dataset(
        self,
        data: pd.DataFrame,
        scorer: str,
        relnames: list[str],
        pointnames: list[str],
        output_dir: Path,
    ) -> None:
        """
        Create complete DeepLabCut dataset with DataFrame and save files.

        Builds the multi-index DataFrame structure and saves to both
        HDF5 and CSV formats expected by DeepLabCut.

        Args:
            data: DataFrame with raw coordinate data
            scorer: Name of scorer/experimenter
            relnames: List of absolute image file paths
            pointnames: List of body part names
            output_dir: Directory to save output files
        """
        # Build the DataFrame
        dataFrame = self.build_dlc_dataframe(data, scorer, relnames, pointnames)

        # Save files
        h5_save_path = output_dir / f"CollectedData_{scorer}.h5"
        csv_save_path = output_dir / f"CollectedData_{scorer}.csv"

        dataFrame.to_hdf(h5_save_path, key="df_with_missing", mode="w")
        dataFrame.to_csv(csv_save_path, na_rep="NaN")

    def read_trial_csv_with_validation(
        self,
        trialnames: list[Path],
    ) -> tuple[list[pd.DataFrame], list[list[int]], list[str]]:
        """
        Read trial CSVs and validate point name consistency.

        Args:
            trialnames: List of trial directory paths

        Returns:
            Tuple of (dataframes, valid_frame_indices, pointnames)
            - dataframes: List of DataFrames (one per trial, header row removed)
            - valid_frame_indices: List of lists containing valid frame indices per trial
            - pointnames: List of point/marker names

        Raises:
            ValueError: If CSVs have inconsistent point names or structure
            FileNotFoundError: If CSV file is missing or multiple CSVs found
        """

        dfs = []
        idx = []
        pnames = []

        for trial in trialnames:
            # Find CSV file
            contents = list(trial.glob("*.csv"))
            if len(contents) != 1:
                csv_list = [f.name for f in contents]
                raise FileNotFoundError(
                    f"Expected exactly 1 CSV file in {trial.name}, "
                    f"but found {len(contents)}: {csv_list}"
                )
            filename = contents[0]

            # Read CSV with header=None to match xrommtools pattern
            df1 = pd.read_csv(filename, sep=",", header=None)

            # Extract point names from header row (every 4th column, strip suffix)
            pointnames = df1.loc[0, ::4].astype(str).str[:-7].tolist()
            pnames.append(pointnames)

            # Remove header row
            df1 = df1.loc[1:,].reset_index(drop=True)

            # Find valid frames (rows where >= 50% of columns are non-NaN)
            ncol = df1.shape[1]
            temp_idx = list(df1.index.values[(~pd.isnull(df1)).sum(axis=1) >= ncol / 2])

            # Randomize frames within each trial
            random.shuffle(temp_idx)
            idx.append(temp_idx)
            dfs.append(df1)

        # Validate consistent point names across trials
        if any(pnames[0] != x for x in pnames):
            # Build detailed error message showing differences
            differences = []
            for i, pname_list in enumerate(pnames):
                if pname_list != pnames[0]:
                    trial_name = trialnames[i].name
                    diff_points = set(pname_list) ^ set(pnames[0])
                    differences.append(f"  {trial_name}: differs by {diff_points}")

            error_msg = (
                "Point names are not consistent across trials.\n"
                f"Reference trial ({trialnames[0].name}): {pnames[0]}\n"
                "Differences found in:\n" + "\n".join(differences)
            )
            raise ValueError(error_msg)

        return dfs, idx, pnames[0]

    def extract_2d_points_for_camera(
        self,
        trial_csv_df: pd.DataFrame,
        camera: int,
        picked_frames: list[int],
    ) -> pd.DataFrame:
        """
        Extract 2D point data for specific camera from trial CSV.

        Args:
            trial_csv_df: DataFrame loaded from trial CSV (header row already removed)
            camera: Camera number (1 or 2)
            picked_frames: List of frame indices to extract

        Returns:
            DataFrame with x,y coordinates for each point, sorted by column index
            Note: Columns are NOT renamed - caller should handle column naming

        Raises:
            ValueError: If camera number is not 1 or 2
            IndexError: If picked_frames contains invalid indices
        """
        # Validate camera number
        if camera not in [1, 2]:
            raise ValueError(f"Camera must be 1 or 2, got {camera}")

        # Validate frame indices
        max_frame = trial_csv_df.shape[0] - 1
        invalid_frames = [f for f in picked_frames if f > max_frame or f < 0]
        if invalid_frames:
            raise IndexError(
                f"Frame indices {invalid_frames} are out of bounds. "
                f"Valid range: 0-{max_frame}"
            )

        # Extract x and y positions for this camera
        # Camera 1: columns 0, 1, 4, 5, 8, 9, ... (offset 0)
        # Camera 2: columns 2, 3, 6, 7, 10, 11, ... (offset 2)
        xpos = trial_csv_df.iloc[picked_frames, 0 + (camera - 1) * 2 :: 4]
        ypos = trial_csv_df.iloc[picked_frames, 1 + (camera - 1) * 2 :: 4]

        # Concatenate and sort by column index
        temp_data = pd.concat([xpos, ypos], axis=1).sort_index(axis=1)

        return temp_data



"""Utility for creating or loading project configs."""

import warnings

import cv2
import deeplabcut
import numpy as np
import pandas as pd
from pathlib import Path
import yaml

from .xma_data_processor import XMADataProcessor

# Sets the default codec for use in creating/loading project configs
DEFAULT_CODEC = "avc1"


def _migrate_tracking_mode(config: dict):
    """Migrate deprecated 'tracking_mode' to 'mode' with backwards compatibility.

    This function handles the transition from 'tracking_mode' to 'mode' introduced
    in version 0.2.5. Support for 'tracking_mode' will be removed in version 1.0.

    Args:
        config: Dictionary containing project configuration

    Returns:
        Modified config dictionary with migration applied

    Raises:
        ValueError: If both keys exist with conflicting values

    Note:
        Support for 'tracking_mode' will be removed in version 1.0.
    """
    has_mode = "mode" in config
    has_tracking_mode = "tracking_mode" in config

    if has_mode and has_tracking_mode:
        if config["mode"] != config["tracking_mode"]:
            raise ValueError(
                f"Conflicting values detected in config: 'mode' is set to "
                f"'{config['mode']}' but 'tracking_mode' is set to "
                f"'{config['tracking_mode']}'. Please remove the deprecated "
                f"'tracking_mode' key from your config and use only 'mode'."
            )
        else:
            warnings.warn(
                "Config contains both 'mode' and 'tracking_mode' with the same "
                "value. Removing deprecated 'tracking_mode'. This key will be "
                "removed in version 1.0.",
                DeprecationWarning,
                stacklevel=3,
            )
            del config["tracking_mode"]
    elif has_tracking_mode:
        warnings.warn(
            "'tracking_mode' is deprecated and will be removed in version 1.0. "
            "Use 'mode' instead. Automatically migrating your config...",
            DeprecationWarning,
            stacklevel=3,
        )
        config["mode"] = config["tracking_mode"]
        del config["tracking_mode"]

    return config


class Project:
    def __init__(self):
        raise NotImplementedError("Use create_new_config or load_config instead.")

    @staticmethod
    def create_new_config(
        working_dir=None, experimenter="NA", mode="2D", codec=DEFAULT_CODEC
    ):
        """Creates a new config from scratch."""
        try:
            working_dir = Path(working_dir) if working_dir is not None else Path.cwd()
            (working_dir / "trainingdata").mkdir(parents=True)
            (working_dir / "trials").mkdir(parents=True)
        except FileExistsError as e:
            print(
                "It looks like this project folder already exists. Try importing it with DeepXROMM.load_project(working_dir)"
            )
            raise e

        # Create a dummy data processor to validate user's codec input
        _data_processor = XMADataProcessor(
            {"swapped_markers": False, "crossed_markers": False}
        )
        _data_processor.validate_codec(codec)

        # Create a fake video to pass into the deeplabcut workflow
        dummy_video_path = working_dir / "dummy.avi"
        frame = np.zeros((480, 480, 3), dtype=np.uint8)
        out = cv2.VideoWriter(
            str(dummy_video_path), cv2.VideoWriter_fourcc(*codec), 15, (480, 480)
        )
        out.write(frame)
        out.release()

        # Create a new project
        task = working_dir.name
        path_config_file = deeplabcut.create_new_project(
            task,
            experimenter,
            [str(dummy_video_path)],
            str(working_dir / ""),  # Add the trailing slash
            copy_videos=True,
        )

        config_path = Path(__file__).parent / "default_config.yaml"
        with config_path.open() as file:
            config_data = yaml.safe_load(file)

        config_data.update(
            {
                "task": task,
                "experimenter": experimenter,
                "working_dir": str(working_dir),
                "path_config_file": path_config_file,
                "mode": mode,
                "video_codec": codec,
            }
        )

        if mode == "per_cam":
            task_2 = f"{task}_cam2"
            path_config_file_2 = deeplabcut.create_new_project(
                task_2,
                experimenter,
                [str(dummy_video_path)],
                str(working_dir / ""),  # Add the trailing slash
                copy_videos=True,
            )
            config_data["path_config_file_2"] = path_config_file_2

        # Save configuration
        config_file = working_dir / "project_config.yaml"
        with config_file.open("w") as config:
            yaml.dump(config_data, config, sort_keys=False)

        # Cleanup
        try:
            (Path(path_config_file).parent / "labeled-data/dummy").rmdir()
        except FileNotFoundError:
            pass

        try:
            (Path(path_config_file).parent / "videos/dummy.avi").unlink()
        except FileNotFoundError:
            pass

        try:
            dummy_video_path.unlink()
        except FileNotFoundError:
            pass
        return config_data

    @staticmethod
    def load_config(working_dir=None):
        """Load an existing project"""
        working_dir = Path(working_dir) if working_dir is not None else Path.cwd()

        # Open the config
        config_path = working_dir / "project_config.yaml"
        with config_path.open("r") as config_file:
            project = yaml.safe_load(config_file)

        # Migrate deprecated 'tracking_mode' to 'mode'
        project = _migrate_tracking_mode(project)

        experimenter = str(project["experimenter"])
        project["experimenter"] = experimenter
        if project["dataset_name"] == "MyData":
            warnings.warn("Default project name in use", SyntaxWarning)

        # Initiate data processor utility
        data_processor = XMADataProcessor(config=project)
        training_trials = data_processor.list_trials("trainingdata")
        if len(training_trials) == 0:
            raise FileNotFoundError(
                "Empty trials directory found. Expected trial folders within the 'trainingdata' directory"
            )
        trial_path = training_trials[0]

        # Load trial CSV
        trial_csv_path = data_processor.find_trial_csv(trial_path)
        trial_csv = pd.read_csv(trial_csv_path)

        # Give search_area a minimum of 10
        project["search_area"] = (
            int(project["search_area"] + 0.5) if project["search_area"] >= 10 else 10
        )

        # Drop untracked frames (all NaNs)
        trial_csv = trial_csv.dropna(how="all")

        # Make sure there aren't any partially tracked frames
        if trial_csv.isna().sum().sum() > 0:
            raise AttributeError(
                f"Detected {len(trial_csv) - len(trial_csv.dropna())} partially tracked frames. \
        Please ensure that all frames are completely tracked"
            )

        # Check/set the default value for tracked frames
        if project["nframes"] <= 0:
            project["nframes"] = len(trial_csv)

        elif project["nframes"] != len(trial_csv):
            warnings.warn(
                "Project nframes tracked does not match 2D Points file. \
            If this is intentional, ignore this message"
            )

        # Check the current nframes against the threshold value * the number of frames in the cam1 video
        cam1_video_path = data_processor.find_cam_file(trial_path, "cam1")
        video = cv2.VideoCapture(cam1_video_path)

        if (
            project["nframes"]
            < int(video.get(cv2.CAP_PROP_FRAME_COUNT)) * project["tracking_threshold"]
        ):
            tracking_threshold = project["tracking_threshold"]
            warnings.warn(
                f"Project nframes is less than the recommended {tracking_threshold * 100}% of the total frames"
            )

        # Check DLC bodyparts (marker names)
        default_bodyparts = ["bodypart1", "bodypart2", "bodypart3", "objectA"]
        bodyparts = data_processor.get_bodyparts_from_xma(
            trial_csv_path, mode=project["mode"]
        )

        # TODO: use DLC's config loader functions to handle this
        dlc_config_path = Path(project["path_config_file"])
        with dlc_config_path.open("r") as dlc_config:
            dlc_yaml = yaml.safe_load(dlc_config)

        if dlc_yaml["bodyparts"] == default_bodyparts:
            dlc_yaml["bodyparts"] = bodyparts
        elif dlc_yaml["bodyparts"] != bodyparts:
            raise SyntaxError(
                "XMAlab CSV marker names are different than DLC bodyparts."
            )

        with dlc_config_path.open("w") as dlc_config:
            yaml.dump(dlc_yaml, dlc_config, sort_keys=False)

        # Check DLC bodyparts (marker names) for config 2 if needed
        if project["mode"] == "per_cam":
            try:
                dlc_config_path_2 = Path(project["path_config_file_2"])
            except KeyError as e:
                print(
                    "Path to second DLC config not found. Did you create the project as a per-cam project?"
                )
                print("If not, re-run 'create_new_project' using mode='per_cam'")
                raise e
            with dlc_config_path_2.open("r") as dlc_config:
                dlc_yaml = yaml.safe_load(dlc_config)
            # Better conditional logic could definitely be had to reduce function calls here
            if dlc_yaml["bodyparts"] == default_bodyparts:
                dlc_yaml["bodyparts"] = bodyparts
            elif dlc_yaml["bodyparts"] != bodyparts:
                raise SyntaxError(
                    "XMAlab CSV marker names are different than DLC bodyparts."
                )

            with dlc_config_path_2.open("w") as dlc_config:
                yaml.dump(dlc_yaml, dlc_config, sort_keys=False)

        # Check test_autocorrect params for defaults
        if project["test_autocorrect"]:
            if project["trial_name"] == "your_trial_here":
                raise SyntaxError("Please specify a trial to test autocorrect() with")
            if project["marker"] == "your_marker_here":
                raise SyntaxError("Please specify a marker to test autocorrect() with")

        # Auto-add default video_codec to config if not set
        if "video_codec" not in project:
            project["video_codec"] = DEFAULT_CODEC

        # Update changed attributes to match in the file
        with config_path.open("w") as file:
            yaml.dump(project, file, sort_keys=False)

        return project



"""
Test elements of the trial process that do not vary between modes
"""

import os
from pathlib import Path
import shutil
import unittest

import cv2
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

from deepxromm import DeepXROMM

SAMPLE_FRAME = Path(__file__).parent / "sample_frame.jpg"
SAMPLE_FRAME_INPUT = Path(__file__).parent / "sample_frame_input.csv"
SAMPLE_AUTOCORRECT_OUTPUT = Path(__file__).parent / "sample_autocorrect_output.csv"

DEEPXROMM_TEST_CODEC = os.environ.get("DEEPXROMM_TEST_CODEC", "avc1")


class TestSampleFrame(unittest.TestCase):
    """Test function behaviors using a frame from an actual trial"""

    def setUp(self):
        """Create trial"""
        self.working_dir = Path.cwd() / "tmp"
        self.deepxromm = DeepXROMM.create_new_project(
            self.working_dir, codec=DEEPXROMM_TEST_CODEC
        )
        frame = cv2.imread(str(SAMPLE_FRAME))

        # Make a trial directory
        trial_dir = self.working_dir / "trainingdata/test"
        trial_dir.mkdir(parents=True, exist_ok=True)

        # Move sample frame input to trainingdata
        shutil.copy(str(SAMPLE_FRAME_INPUT), str(trial_dir / "test.csv"))

        # Cam 1 and Cam 2 trainingdata setup
        for cam in ["cam1", "cam2"]:
            video_path = trial_dir / f"dummy_{cam}.avi"
            out = cv2.VideoWriter(
                str(video_path), cv2.VideoWriter_fourcc(*"DIVX"), 30, (1024, 512)
            )
            out.write(frame)
            out.release()

        # Move sample frame input to trials (it0 and trials)
        trials_dir = self.working_dir / "trials/test/it0"
        trials_dir.mkdir(parents=True, exist_ok=True)
        shutil.copy(
            str(SAMPLE_FRAME_INPUT), str(self.working_dir / "trials/test/test.csv")
        )
        shutil.copy(
            str(SAMPLE_FRAME_INPUT), str(trials_dir / "test-Predicted2DPoints.csv")
        )

        # Cam 1 and Cam 2 trials setup
        for cam in ["cam1", "cam2"]:
            video_path = self.working_dir / f"trials/test/test_{cam}.avi"
            out = cv2.VideoWriter(
                str(video_path), cv2.VideoWriter_fourcc(*"DIVX"), 30, (1024, 512)
            )
            out.write(frame)
            out.release()

        cv2.destroyAllWindows()

    def test_autocorrect_always_uses_pred_points_csv(self):
        """Test that autocorrect always uses predicted points file"""

        # Delete other CSVs to ensure autocorrect is only running against the it0 CSV
        # Main trial CSV
        test_trial_csv = Path(self.working_dir / "trials/test/test.csv")
        test_trial_csv.unlink()

        # Training data CSV
        training_data_csv = Path(self.working_dir / "trainingdata/test/test.csv")
        training_data_csv.unlink()

        # Run autocorrect
        self.deepxromm.autocorrect_trials()

        # Copy CSVs back in to make sure other tests pass (better way to do?)
        new_trial_csv = Path(self.working_dir / "trials/test/test.csv")
        new_training_data_csv = Path(self.working_dir / "trainingdata/test/test.csv")
        shutil.copy(str(SAMPLE_FRAME_INPUT), str(new_trial_csv))
        shutil.copy(str(SAMPLE_FRAME_INPUT), str(new_training_data_csv))

    def test_autocorrect_does_correction(self):
        """Make sure that autocorrect corrects the frame after making changes"""
        # Run autocorrect on the sample frame
        self.deepxromm.autocorrect_trials()

        # Load CSVs
        function_output_path = (
            self.working_dir / "trials/test/it0/test-AutoCorrected2DPoints.csv"
        )
        function_output = pd.read_csv(str(function_output_path), dtype="float64")
        sample_output = pd.read_csv(str(SAMPLE_AUTOCORRECT_OUTPUT), dtype="float64")

        # Drop cam2 markers and check for changes
        columns_to_drop = function_output.columns[
            function_output.columns.str.contains("cam2", case=False)
        ]
        function_output.drop(columns_to_drop, axis=1, inplace=True)
        function_output = function_output.round(6)
        sample_output = sample_output.round(6)

        try:
            # Use assert_frame_equal to check if the data frames are the same
            assert_frame_equal(
                function_output, sample_output, check_exact=False, rtol=1e-6, atol=1e-6
            )
        except AssertionError as e:
            print(f"Autocorrector diff: {function_output.compare(sample_output)}")
            raise e

    def test_image_hashing_identical_trials_returns_0(self):
        """Make sure the image hashing function is working properly"""
        # Create an identical second trial
        frame_path = self.working_dir.parent / "sample_frame.jpg"
        frame = cv2.imread(str(frame_path))
        trial_dir = self.working_dir / "trials/test2_same"
        trial_dir.mkdir(parents=True, exist_ok=True)

        # Create videos
        for cam in ["cam1", "cam2"]:
            video_path = trial_dir / f"test2_same_{cam}.avi"
            out = cv2.VideoWriter(
                str(video_path), cv2.VideoWriter_fourcc(*"DIVX"), 30, (1024, 512)
            )
            out.write(frame)
            out.release()

        # Do similarity comparison
        similarity = self.deepxromm.analyze_video_similarity_project()
        # Since both videos are the same, the image similarity output should be 0
        self.assertEqual(sum(similarity.values()), 0)

    def test_image_hashing_different_trials_returns_nonzero(self):
        """Image hashing different videos returns nonzero answer"""
        # Create a different second trial
        frame = np.zeros((480, 480, 3), np.uint8)
        trial_dir = self.working_dir / "trials/test2_diff"
        trial_dir.mkdir(parents=True, exist_ok=True)

        # Create videos
        for cam in ["cam1", "cam2"]:
            video_path = trial_dir / f"test2_diff_{cam}.avi"
            out = cv2.VideoWriter(
                str(video_path), cv2.VideoWriter_fourcc(*"DIVX"), 30, (480, 480)
            )
            out.write(frame)
            out.release()

        # Do similarity comparison
        similarity = self.deepxromm.analyze_video_similarity_project()
        # Since the videos are different, should return nonzero answer
        self.assertNotEqual(sum(similarity.values()), 0)

    def test_marker_similarity_returns_0_if_identical(self):
        """Check that identical data has a similarity value of 0"""
        # Move sample data into test trial
        shutil.copy(
            str(SAMPLE_FRAME_INPUT), str(self.working_dir / "trials/test/test.csv")
        )

        # Move sample data into test2 trial
        (self.working_dir / "trials/test2").mkdir(parents=True, exist_ok=True)
        shutil.copy(
            str(SAMPLE_FRAME_INPUT), str(self.working_dir / "trials/test2/test2.csv")
        )

        # Do cross-correlation
        marker_similarity = self.deepxromm.analyze_marker_similarity_project()
        self.assertEqual(sum(marker_similarity.values()), 0)

    def test_marker_similarity_returns_not_0_if_different(self):
        """Check that different data has a similarity value of not 0"""
        # Move sample data into test trial
        shutil.copy(
            str(SAMPLE_FRAME_INPUT), str(self.working_dir / "trials/test/test.csv")
        )

        # Move autocorrect data into test2 trial
        (self.working_dir / "trials/test2").mkdir(parents=True, exist_ok=True)
        shutil.copy(
            str(SAMPLE_AUTOCORRECT_OUTPUT),
            str(self.working_dir / "trials/test2/test2.csv"),
        )

        # Do cross-correlation
        marker_similarity = self.deepxromm.analyze_marker_similarity_project()
        self.assertNotEqual(sum(marker_similarity.values()), 0)

    def tearDown(self):
        """Remove the created temp project"""
        project_path = Path.cwd() / "tmp"
        shutil.rmtree(project_path)
